# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from functools import partial

import pytest
import scipp as sc
import tof as tof_pkg

from scippneutron.chopper import DiskChopper
from scippneutron.conversion.graph.beamline import beamline
from scippneutron.conversion.graph.tof import elastic
from scippneutron.tof import chopper_cascade, fakes, unwrap

sl = pytest.importorskip('sciline')


@pytest.fixture
def disk_choppers():
    psc1 = DiskChopper(
        frequency=sc.scalar(14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(286 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.145], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=[-1.23, 70.49, 84.765, 113.565, 170.29, 271.635, 286.035, 301.17],
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=[1.23, 73.51, 88.035, 116.835, 175.31, 275.565, 289.965, 303.63],
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    psc2 = DiskChopper(
        frequency=sc.scalar(-14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-236, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.155], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=[-1.23, 27.0, 55.8, 142.385, 156.765, 214.115, 257.23, 315.49],
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=[1.23, 30.6, 59.4, 145.615, 160.035, 217.885, 261.17, 318.11],
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    oc = DiskChopper(
        frequency=sc.scalar(14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(297 - 180 - 90, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.174], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-27.6 * 0.5], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[27.6 * 0.5], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    bcc = DiskChopper(
        frequency=sc.scalar(112.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(240 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 9.78], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-36.875, 143.125], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[36.875, 216.875], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    t0 = DiskChopper(
        frequency=sc.scalar(28.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(280 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 13.05], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-314.9 * 0.5], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[314.9 * 0.5], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    return {"psc1": psc1, "psc2": psc2, "oc": oc, "bcc": bcc, "t0": t0}


@pytest.fixture
def overlap_choppers(disk_choppers):
    out = disk_choppers.copy()
    out['bcc'] = DiskChopper(
        frequency=sc.scalar(112.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(240 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 9.78], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-36.875, 143.125], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[46.875, 216.875], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )
    return out


@pytest.mark.parametrize("npulses", [1, 2])
@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=['detector_number'], values=[77.675], unit='m'),
        sc.array(dims=['detector_number'], values=[77.675, 76.5], unit='m'),
        sc.array(
            dims=['y', 'x'],
            values=[[77.675, 76.1, 78.05], [77.15, 77.3, 77.675]],
            unit='m',
        ),
    ],
)
def test_dream_wfm(disk_choppers, npulses, ltotal):
    choppers = {
        key: chopper_cascade.Chopper.from_disk_chopper(
            chop, pulse_frequency=sc.scalar(14.0, unit="Hz"), npulses=npulses
        )
        for key, chop in disk_choppers.items()
    }

    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to='detector'))
    }

    # Create some neutron events
    wavelengths = sc.array(
        dims=['event'], values=[1.5, 1.6, 1.7, 3.3, 3.4, 3.5], unit='angstrom'
    )
    birth_times = sc.full(sizes=wavelengths.sizes, value=1.5, unit='ms')
    ess_beamline = fakes.FakeBeamlineEss(
        choppers=choppers,
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * npulses,
        events_per_pulse=len(wavelengths),
        source=partial(
            tof_pkg.Source.from_neutrons,
            birth_times=birth_times,
            wavelengths=wavelengths,
            frequency=sc.scalar(14.0, unit="Hz"),
        ),
    )

    # Save the true wavelengths for later
    true_wavelengths = ess_beamline.source.data.coords["wavelength"]

    raw_data = sc.concat(
        [ess_beamline.get_monitor(key) for key in monitors.keys()],
        dim='detector',
    ).fold(dim='detector', sizes=ltotal.sizes)

    # Verify that all 6 neutrons made it through the chopper cascade
    assert sc.identical(
        raw_data.bins.concat('pulse').hist().data,
        sc.array(
            dims=['detector'],
            values=[len(wavelengths) * npulses] * len(monitors),
            unit="counts",
            dtype='float64',
        ).fold(dim='detector', sizes=ltotal.sizes),
    )

    # Set up the workflow
    workflow = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers(wfm=True)
    )
    workflow[unwrap.PulsePeriod] = sc.reciprocal(ess_beamline.source.frequency)
    workflow[unwrap.PulseStride | None] = None

    # Define the extent of the pulse that contains the 6 neutrons in time and wavelength
    # Note that we make a larger encompassing pulse to ensure that the frame bounds are
    # computed correctly
    workflow[unwrap.SourceTimeRange] = (
        sc.scalar(0.0, unit='ms'),
        sc.scalar(4.9, unit='ms'),
    )
    workflow[unwrap.SourceWavelengthRange] = (
        sc.scalar(0.2, unit='angstrom'),
        sc.scalar(16.0, unit='angstrom'),
    )

    workflow[unwrap.Choppers] = choppers
    workflow[unwrap.Ltotal] = ltotal
    workflow[unwrap.RawData] = raw_data

    # Compute time-of-flight
    tofs = workflow.compute(unwrap.TofData)
    assert {dim: tofs.sizes[dim] for dim in ltotal.sizes} == ltotal.sizes

    # Convert to wavelength
    graph = {**beamline(scatter=False), **elastic("tof")}
    wav_wfm = tofs.transform_coords("wavelength", graph=graph)

    # Compare the computed wavelengths to the true wavelengths
    for i in range(npulses):
        result = wav_wfm['pulse', i].flatten(to='detector')
        for j in range(len(result)):
            computed_wavelengths = result[j].values.coords["wavelength"]
            assert sc.allclose(
                computed_wavelengths,
                true_wavelengths['pulse', i],
                rtol=sc.scalar(1e-02),
            )


@pytest.mark.parametrize("npulses", [1, 2])
@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=['detector_number'], values=[77.675], unit='m'),
        sc.array(dims=['detector_number'], values=[77.675, 76.5], unit='m'),
        sc.array(
            dims=['y', 'x'],
            values=[[77.675, 76.1, 78.05], [77.15, 77.3, 77.675]],
            unit='m',
        ),
    ],
)
def test_dream_wfm_with_subframe_time_overlap(overlap_choppers, npulses, ltotal):
    choppers = {
        key: chopper_cascade.Chopper.from_disk_chopper(
            chop, pulse_frequency=sc.scalar(14.0, unit="Hz"), npulses=npulses
        )
        for key, chop in overlap_choppers.items()
    }

    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to='detector'))
    }

    # Create some neutron events
    wavelengths = [1.5, 1.6, 1.7, 3.3, 3.4, 3.5]
    birth_times = [1.5] * len(wavelengths)

    # Add overlap neutrons
    birth_times.extend([0.0, 3.1])
    wavelengths.extend([2.7, 2.5])

    wavelengths = sc.array(dims=['event'], values=wavelengths, unit='angstrom')
    birth_times = sc.array(dims=['event'], values=birth_times, unit='ms')

    ess_beamline = fakes.FakeBeamlineEss(
        choppers=choppers,
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * npulses,
        events_per_pulse=len(wavelengths),
        source=partial(
            tof_pkg.Source.from_neutrons,
            birth_times=birth_times,
            wavelengths=wavelengths,
            frequency=sc.scalar(14.0, unit="Hz"),
        ),
    )

    # Save the true wavelengths for later
    true_wavelengths = ess_beamline.source.data.coords["wavelength"]

    raw_data = sc.concat(
        [ess_beamline.get_monitor(key) for key in monitors.keys()],
        dim='detector',
    ).fold(dim='detector', sizes=ltotal.sizes)

    # Verify that all 6 neutrons made it through the chopper cascade
    assert sc.identical(
        raw_data.bins.concat('pulse').hist().data,
        sc.array(
            dims=['detector'],
            values=[len(wavelengths) * npulses] * len(monitors),
            unit="counts",
            dtype='float64',
        ).fold(dim='detector', sizes=ltotal.sizes),
    )

    # Set up the workflow
    workflow = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers(wfm=True)
    )
    workflow[unwrap.PulsePeriod] = sc.reciprocal(ess_beamline.source.frequency)
    workflow[unwrap.PulseStride | None] = None

    # Define the extent of the pulse that contains the 6 neutrons in time and wavelength
    # Note that we make a larger encompassing pulse to ensure that the frame bounds are
    # computed correctly
    workflow[unwrap.SourceTimeRange] = (
        sc.scalar(0.0, unit='ms'),
        sc.scalar(4.9, unit='ms'),
    )
    workflow[unwrap.SourceWavelengthRange] = (
        sc.scalar(0.2, unit='angstrom'),
        sc.scalar(16.0, unit='angstrom'),
    )

    workflow[unwrap.Choppers] = choppers
    workflow[unwrap.Ltotal] = ltotal
    workflow[unwrap.RawData] = raw_data

    # Compute time-of-flight
    tofs = workflow.compute(unwrap.TofData)
    assert {dim: tofs.sizes[dim] for dim in ltotal.sizes} == ltotal.sizes

    # Convert to wavelength
    graph = {**beamline(scatter=False), **elastic("tof")}
    wav_wfm = tofs.transform_coords("wavelength", graph=graph)

    # Compare the computed wavelengths to the true wavelengths
    for i in range(npulses):
        result_wav = wav_wfm['pulse', i].flatten(to='detector')
        result_tof = tofs['pulse', i].flatten(to='detector')
        for j in range(len(result_wav)):
            computed_tofs = result_tof[j].values.coords["tof"]
            # The two neutrons in the overlap region should have NaN tofs
            assert sc.isnan(computed_tofs[-2])
            assert sc.isnan(computed_tofs[-1])

            computed_wavelengths = result_wav[j].values.coords["wavelength"]
            assert sc.allclose(
                computed_wavelengths[:-2],
                true_wavelengths['pulse', i][:-2],
                rtol=sc.scalar(1e-02),
            )
            # The two neutrons in the overlap region should have NaN wavelengths
            assert sc.isnan(computed_wavelengths[-2])
            assert sc.isnan(computed_wavelengths[-1])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import matplotlib.pyplot as plt
import scipp as sc

from scippneutron.tof import TimeDistanceDiagram


def default_frame_diagram(tmax=None):
    fig, ax = plt.subplots(1, 1)
    diagram = TimeDistanceDiagram(
        ax, tmax=300 * sc.Unit('ms') if tmax is None else tmax
    )
    det1 = 32.0 * sc.Unit('m')
    det2 = 40.0 * sc.Unit('m')
    lambda_min = 16.0 * sc.units.angstrom
    frame_offset = diagram.to_time(6.0 * sc.Unit('ms'))
    diagram.add_source_pulse(15 * sc.Unit('ms'))
    diagram.add_sample(distance=20.0 * sc.Unit('m'))
    diagram.add_detector(distance=det1, name='detector1')
    diagram.add_detector(distance=det2, name='detector2')
    diagram.add_neutrons(
        Lmax=1.05 * det2, lambda_min=lambda_min, time_offset=frame_offset
    )
    diagram.add_neutron(
        L=det1,
        time_offset=frame_offset,
        wavelength=20.0 * sc.units.angstrom,
        label=r'$T_0^{i+1}+\Delta t$',
    )

    props = {'arrowstyle': '-|>'}
    x0 = diagram.frame_length
    x1 = diagram.frame_length + frame_offset
    m = sc.Unit('m')
    ms = sc.Unit('ms')
    diagram.annotate(
        r'$T_0^i$', xy=(x0, 0 * m), xytext=(x0 - 20 * ms, 3 * m), arrowprops=props
    )
    diagram.annotate(
        r'$T_0^i+\Delta T_0$',
        xy=(x1, 0 * m),
        xytext=(x1 - 20 * ms, 5 * m),
        arrowprops=props,
    )
    diagram.annotate(
        r'$T_0^{i+1}+t_{\mathrm{pivot}}(\mathrm{det1})$',
        xy=(205 * ms, det1),
        xytext=(150 * ms, 10 * m),
        arrowprops=props,
    )
    diagram.annotate(
        r'$T_0^{i+2}+t_{\mathrm{pivot}}(\mathrm{det2})$',
        xy=(235 * ms, det2),
        xytext=(220 * ms, 15 * m),
        arrowprops=props,
    )

    # Pivot line
    diagram.add_neutron(
        time_offset=x1, L=1.1 * det2, wavelength=lambda_min, color='black', ls='dashed'
    )

    L = diagram.frame_length.value
    ax.axvspan(2 * L, 3 * L, facecolor='grey', alpha=0.2)

    return fig


def frame_skipping_diagram(tmax=None):
    fig, ax = plt.subplots(1, 1)
    diagram = TimeDistanceDiagram(
        ax, tmax=500 * sc.Unit('ms') if tmax is None else tmax
    )
    det1 = 32.0 * sc.Unit('m')
    det2 = 40.0 * sc.Unit('m')
    lambda_min = 16.0 * sc.units.angstrom
    frame_offset = diagram.to_time(1.5 * sc.Unit('ms'))
    diagram.add_source_pulse()
    diagram.add_sample(distance=20.0 * sc.Unit('m'))
    diagram.add_detector(distance=det1, name='detector1')
    diagram.add_detector(distance=det2, name='detector2')
    diagram.add_neutrons(
        Lmax=1.05 * det2, lambda_min=lambda_min, time_offset=frame_offset, stride=2
    )

    props = {'arrowstyle': '-|>'}
    x0 = 2 * diagram.frame_length
    x1 = 2 * diagram.frame_length + frame_offset
    m = sc.Unit('m')
    ms = sc.Unit('ms')
    diagram.annotate(
        r'$T_0^i$', xy=(x0, 0 * m), xytext=(x0 - 20 * ms, 3 * m), arrowprops=props
    )
    diagram.annotate(
        r'$T_0^i+\Delta T_0$', xy=(x1, 0 * m), xytext=(x1, 5 * m), arrowprops=props
    )

    return fig



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import math

import numpy as np
import pytest
import scipp as sc
import scipp.testing

from scippneutron.peaks import model


@pytest.mark.parametrize('prefix', ['', 'lorentz_'])
def test_polynomial_degree_1_call(prefix: str):
    a0 = sc.scalar(0.2, unit='cm')
    a1 = sc.scalar(-5.1, unit='cm/s')
    params = {f'{prefix}a0': a0, f'{prefix}a1': a1}
    m = model.PolynomialModel(degree=1, prefix=prefix)

    x = sc.linspace('xx', 0.8, 4.3, 100, unit='s')
    expected = a0 + a1 * x
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize('prefix', ['', 'lorentz_'])
def test_polynomial_degree_2_call(prefix: str):
    a0 = sc.scalar(-1.3, unit='K')
    a1 = sc.scalar(0.03, unit='K/m')
    a2 = sc.scalar(1.7, unit='K/m^2')
    params = {f'{prefix}a0': a0, f'{prefix}a1': a1, f'{prefix}a2': a2}
    m = model.PolynomialModel(degree=2, prefix=prefix)

    x = sc.linspace('xx', 100.0, 234.0, 100, unit='m')
    expected = a0 + a1 * x + a2 * x**2
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize('prefix', ['', 'lorentz_'])
def test_polynomial_degree_3_call(prefix: str):
    a0 = sc.scalar(-1.3, unit='K')
    a1 = sc.scalar(0.03, unit='K/m')
    a2 = sc.scalar(1.7, unit='K/m^2')
    a3 = sc.scalar(0.9, unit='K/m^3')
    params = {
        f'{prefix}a0': a0,
        f'{prefix}a1': a1,
        f'{prefix}a2': a2,
        f'{prefix}a3': a3,
    }
    m = model.PolynomialModel(degree=3, prefix=prefix)

    x = sc.linspace('xx', 100.0, 234.0, 100, unit='m')
    expected = a0 + a1 * x + a2 * x**2 + a3 * x**3
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize('prefix', ['', 'lorentz_'])
def test_gaussian_call(prefix: str):
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    params = {
        f'{prefix}amplitude': amplitude,
        f'{prefix}loc': loc,
        f'{prefix}scale': scale,
    }
    m = model.GaussianModel(prefix=prefix)

    x = sc.linspace('xx', -2.0, 3.5, 200, unit='m')
    expected = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize('prefix', ['', 'lorentz_'])
def test_lorentzian_call(prefix: str):
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    params = {
        f'{prefix}amplitude': amplitude,
        f'{prefix}loc': loc,
        f'{prefix}scale': scale,
    }

    m = model.LorentzianModel(prefix=prefix)

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    expected = amplitude / np.pi * scale / ((x - loc) ** 2 + scale**2)
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize('prefix', ['', 'lorentz_'])
@pytest.mark.parametrize('fraction', [0.0, 1.0, 0.5, 0.3])
def test_pseudo_voigt_call(prefix: str, fraction: float):
    amplitude = sc.scalar(2.8, unit='kg*m')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    fraction = sc.scalar(fraction)
    params = {
        f'{prefix}amplitude': amplitude,
        f'{prefix}loc': loc,
        f'{prefix}scale': scale,
        f'{prefix}fraction': fraction,
    }

    m = model.PseudoVoigtModel(prefix=prefix)

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    scale_g = scale / math.sqrt(2 * math.log(2))
    gaussian = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale_g)
        * sc.exp(-((x - loc) ** 2) / (2 * scale_g**2))
    )
    lorentzian = amplitude / np.pi * scale / ((x - loc) ** 2 + scale**2)
    expected = fraction * lorentzian + (1 - fraction) * gaussian
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


@pytest.mark.parametrize('prefix', ['', 'pre_'])
def test_composite_call(prefix: str):
    amplitude = sc.scalar(2.8, unit='kg*m')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    a0 = sc.scalar(-0.4, unit='kg')
    a1 = sc.scalar(0.05, unit='kg/m')
    params = {
        f'{prefix}n_amplitude': amplitude,
        f'{prefix}n_loc': loc,
        f'{prefix}n_scale': scale,
        f'{prefix}p_a0': a0,
        f'{prefix}p_a1': a1,
    }
    m = model.CompositeModel(
        model.PolynomialModel(degree=1, prefix='p_'),
        model.GaussianModel(prefix='n_'),
        prefix=prefix,
    )

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    expected_p = a0 + a1 * x
    expected_n = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    expected = expected_p + expected_n
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


def test_polynomial_degree_1_guess_params():
    a0 = sc.scalar(0.2, unit='cm')
    a1 = sc.scalar(-5.1, unit='cm/s')

    x = sc.linspace('xx', 0.8, 4.3, 100, unit='s')
    y = a0 + a1 * x
    data = sc.DataArray(y, coords={'xx': x})

    m = model.PolynomialModel(degree=1, prefix='')
    params = m.guess(data)
    assert params.keys() == {'a0', 'a1'}
    sc.testing.assert_allclose(params['a0'], a0, atol=sc.scalar(0.1, unit='cm'))
    sc.testing.assert_allclose(params['a1'], a1, atol=sc.scalar(0.1, unit='cm/s'))


def test_polynomial_degree_1_guess_params_prefix():
    a0 = sc.scalar(0.2, unit='cm')
    a1 = sc.scalar(-5.1, unit='cm/s')

    x = sc.linspace('xx', 0.8, 4.3, 100, unit='s')
    y = a0 + a1 * x
    data = sc.DataArray(y, coords={'xx': x})

    m = model.PolynomialModel(degree=1, prefix='linear_')
    params = m.guess(data)
    assert params.keys() == {'linear_a1', 'linear_a0'}
    sc.testing.assert_allclose(params['linear_a0'], a0, atol=sc.scalar(0.1, unit='cm'))
    sc.testing.assert_allclose(
        params['linear_a1'], a1, atol=sc.scalar(0.1, unit='cm/s')
    )


def test_polynomial_degree_2_guess_params():
    a0 = sc.scalar(-1.3, unit='K')
    a1 = sc.scalar(0.03, unit='K/m')
    a2 = sc.scalar(1.7, unit='K/m^2')

    x = sc.linspace('xx', 100.0, 234.0, 100, unit='m')
    y = a0 + a1 * x + a2 * x**2
    data = sc.DataArray(y, coords={'xx': x})

    m = model.PolynomialModel(degree=2, prefix='')
    params = m.guess(data)
    assert params.keys() == {'a0', 'a1', 'a2'}
    sc.testing.assert_allclose(params['a0'], a0, atol=sc.scalar(0.1, unit='K'))
    sc.testing.assert_allclose(params['a1'], a1, atol=sc.scalar(0.1, unit='K/m'))
    sc.testing.assert_allclose(params['a2'], a2, atol=sc.scalar(0.1, unit='K/m^2'))


def test_gaussian_guess_params_linspace():
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    y = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    data = sc.DataArray(y, coords={'xx': x})

    m = model.GaussianModel(prefix='')
    params = m.guess(data)
    assert params.keys() == {'amplitude', 'loc', 'scale'}
    sc.testing.assert_allclose(
        params['amplitude'], amplitude, atol=sc.scalar(0.5, unit='kg')
    )
    sc.testing.assert_allclose(params['loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['scale'], scale, atol=sc.scalar(0.5, unit='m'))


def test_gaussian_guess_params_logspace():
    amplitude = sc.scalar(0.02, unit='kg')
    loc = sc.scalar(10.2, unit='m')
    scale = sc.scalar(2.6, unit='m')

    x = sc.geomspace('xx', 2.0, 20.0, 200, unit='m')
    y = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    data = sc.DataArray(y, coords={'xx': x})

    m = model.GaussianModel(prefix='')
    params = m.guess(data)
    assert params.keys() == {'amplitude', 'loc', 'scale'}
    sc.testing.assert_allclose(
        params['amplitude'], amplitude, atol=sc.scalar(0.05, unit='kg')
    )
    sc.testing.assert_allclose(params['loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['scale'], scale, atol=sc.scalar(0.5, unit='m'))


def test_gaussian_guess_params_disordered():
    rng = np.random.default_rng(772)

    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    y = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    data = sc.DataArray(y, coords={'xx': x})
    data = data[list(rng.permutation(len(data)))]

    m = model.GaussianModel(prefix='')
    params = m.guess(data)
    assert params.keys() == {'amplitude', 'loc', 'scale'}
    sc.testing.assert_allclose(
        params['amplitude'], amplitude, atol=sc.scalar(0.5, unit='kg')
    )
    sc.testing.assert_allclose(params['loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['scale'], scale, atol=sc.scalar(0.5, unit='m'))


def test_gaussian_guess_params_small_array():
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')

    x = sc.linspace('xx', 0.2, 0.5, 2, unit='m')
    y = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    data = sc.DataArray(y, coords={'xx': x})

    m = model.GaussianModel(prefix='')
    params = m.guess(data)
    assert params.keys() == {'amplitude', 'loc', 'scale'}
    sc.testing.assert_allclose(
        params['amplitude'], amplitude, atol=sc.scalar(5.0, unit='kg')
    )
    sc.testing.assert_allclose(params['loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['scale'], scale, atol=sc.scalar(0.5, unit='m'))


def test_gaussian_guess_params_prefix():
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    y = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    data = sc.DataArray(y, coords={'xx': x})

    m = model.GaussianModel(prefix='normal-')
    params = m.guess(data)
    assert params.keys() == {'normal-amplitude', 'normal-loc', 'normal-scale'}
    sc.testing.assert_allclose(
        params['normal-amplitude'], amplitude, atol=sc.scalar(0.5, unit='kg')
    )
    sc.testing.assert_allclose(params['normal-loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(
        params['normal-scale'], scale, atol=sc.scalar(0.5, unit='m')
    )


def test_lorentzian_guess_params():
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    y = amplitude / np.pi * scale / ((x - loc) ** 2 + scale**2)
    data = sc.DataArray(y, coords={'xx': x})

    m = model.LorentzianModel(prefix='')
    params = m.guess(data)
    assert params.keys() == {'amplitude', 'loc', 'scale'}
    sc.testing.assert_allclose(
        params['amplitude'], amplitude, atol=sc.scalar(0.6, unit='kg')
    )
    sc.testing.assert_allclose(params['loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['scale'], scale, atol=sc.scalar(0.5, unit='m'))


@pytest.mark.parametrize('fraction', [0.0, 1.0, 0.5, 0.3])
def test_pseudo_voigt_guess_params(fraction: float):
    amplitude = sc.scalar(2.8, unit='kg')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    fraction = sc.scalar(fraction)

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    scale_g = scale / math.sqrt(2 * math.log(2))
    gaussian = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale_g)
        * sc.exp(-((x - loc) ** 2) / (2 * scale_g**2))
    )
    lorentzian = amplitude / np.pi * scale / ((x - loc) ** 2 + scale**2)
    y = (1 - fraction) * gaussian + fraction * lorentzian
    data = sc.DataArray(y, coords={'xx': x})

    m = model.PseudoVoigtModel(prefix='')
    params = m.guess(data)
    assert params.keys() == {'amplitude', 'loc', 'scale', 'fraction'}
    sc.testing.assert_allclose(
        params['amplitude'], amplitude, atol=sc.scalar(0.6, unit='kg')
    )
    sc.testing.assert_allclose(params['loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['scale'], scale, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_identical(params['fraction'], sc.scalar(0.5))


def test_composite_guess_params():
    amplitude = sc.scalar(2.8, unit='kg*m')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    a0 = sc.scalar(-0.4, unit='kg')
    a1 = sc.scalar(0.05, unit='kg/m')
    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    y_p = a0 + a1 * x
    y_n = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    y = y_p + y_n
    data = sc.DataArray(y, coords={'xx': x})

    m = model.CompositeModel(
        model.PolynomialModel(degree=1, prefix='p_'),
        model.GaussianModel(prefix='n_'),
        prefix='',
    )
    params = m.guess(data)
    assert params.keys() == {'p_a0', 'p_a1', 'n_amplitude', 'n_loc', 'n_scale'}
    sc.testing.assert_allclose(params['p_a0'], a0, atol=sc.scalar(0.5, unit='kg'))
    sc.testing.assert_allclose(params['p_a1'], a1, atol=sc.scalar(0.5, unit='kg/m'))
    sc.testing.assert_allclose(
        params['n_amplitude'], amplitude, atol=sc.scalar(0.6, unit='kg*m')
    )
    sc.testing.assert_allclose(params['n_loc'], loc, atol=sc.scalar(0.5, unit='m'))
    sc.testing.assert_allclose(params['n_scale'], scale, atol=sc.scalar(0.5, unit='m'))


def test_composite_from_sum():
    amplitude = sc.scalar(2.8, unit='kg*m')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    a0 = sc.scalar(-0.4, unit='kg')
    a1 = sc.scalar(0.05, unit='kg/m')
    params = {
        'n_amplitude': amplitude,
        'n_loc': loc,
        'n_scale': scale,
        'p_a0': a0,
        'p_a1': a1,
    }
    m = model.PolynomialModel(degree=1, prefix='p_') + model.GaussianModel(prefix='n_')

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    expected_p = a0 + a1 * x
    expected_n = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    expected = expected_p + expected_n
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


def test_composite_without_prefixes():
    amplitude = sc.scalar(2.8, unit='kg*m')
    loc = sc.scalar(0.4, unit='m')
    scale = sc.scalar(0.1, unit='m')
    a0 = sc.scalar(-0.4, unit='kg')
    a1 = sc.scalar(0.05, unit='kg/m')
    params = {
        'amplitude': amplitude,
        'loc': loc,
        'scale': scale,
        'a0': a0,
        'a1': a1,
    }

    left = model.PolynomialModel(degree=1)
    right = model.GaussianModel()
    m = left + right

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    expected_p = a0 + a1 * x
    expected_n = (
        amplitude
        / (math.sqrt(2 * math.pi) * scale)
        * sc.exp(-((x - loc) ** 2) / (2 * scale**2))
    )
    expected = expected_p + expected_n
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


def test_composite_param_name_clash():
    left = model.PolynomialModel(degree=1, prefix='p_')
    right = model.PolynomialModel(degree=2, prefix='p_')
    with pytest.raises(ValueError, match='overlap'):
        left + right


def test_composite_clash_between_param_and_prefix():
    amplitude1 = sc.scalar(2.8, unit='kg*m')
    loc1 = sc.scalar(0.4, unit='m')
    scale1 = sc.scalar(0.1, unit='m')
    amplitude2 = sc.scalar(1.7, unit='kg*m')
    loc2 = sc.scalar(1.2, unit='m')
    scale2 = sc.scalar(0.2, unit='m')
    params = {
        'amplitude': amplitude1,
        'loc': loc1,
        'scale': scale1,
        'amplitudeamplitude': amplitude2,
        'amplitudeloc': loc2,
        'amplitudescale': scale2,
    }

    left = model.GaussianModel(prefix='')  # has param 'amplitude'
    right = model.GaussianModel(prefix='amplitude')  # has param 'amplitudeamplitude'
    m = left + right

    x = sc.linspace('xx', -4.0, 5.0, 200, unit='m')
    expected1 = (
        amplitude1
        / (math.sqrt(2 * math.pi) * scale1)
        * sc.exp(-((x - loc1) ** 2) / (2 * scale1**2))
    )
    expected2 = (
        amplitude2
        / (math.sqrt(2 * math.pi) * scale2)
        * sc.exp(-((x - loc2) ** 2) / (2 * scale2**2))
    )
    expected = expected1 + expected2
    actual = m(x, **params)
    sc.testing.assert_allclose(actual, expected)


def test_polynomial_param_bounds():
    m = model.PolynomialModel(degree=1)
    assert m.param_bounds == {}


def test_gaussian_param_bounds():
    m = model.GaussianModel()
    assert m.param_bounds == {
        'scale': (0.0, np.inf),
    }


def test_gaussian_param_bounds_prefix():
    m = model.GaussianModel(prefix='g_')
    assert m.param_bounds == {
        'g_scale': (0.0, np.inf),
    }


def test_lorentzian_param_bounds():
    m = model.LorentzianModel()
    assert m.param_bounds == {
        'scale': (0.0, np.inf),
    }


def test_pseudo_voigt_param_bounds():
    m = model.PseudoVoigtModel()
    assert m.param_bounds == {
        'scale': (0.0, np.inf),
        'fraction': (0.0, 1.0),
    }



from abc import ABCMeta, abstractmethod
from typing import Any

import scipp as sc


class SampleShape(metaclass=ABCMeta):
    @abstractmethod
    def beam_intersection(
        self, start_point: sc.Variable, direction: sc.Variable
    ) -> sc.Variable:
        '''Computes the length of the intersection between the shape and the beam
        starting at `start_point` and travelling in the direction `direction`.'''
        pass

    @property
    @abstractmethod
    def volume(self) -> sc.Variable:
        '''Volume of the shape'''
        pass

    @abstractmethod
    def quadrature(self, kind: Any) -> tuple[sc.Variable, sc.Variable]:
        '''Returns quadrature points and weights for evaluating integrals over
        the shape. The method returns a tuple where the first entry is
        an array containing vectors representing points in the shape and the
        second entry is an array containing the weights associated with
        the points.

        Parameters
        -----------
        kind:
            if the shape supports different kinds of quadratures
            this argument denotes which one to use
        '''
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
"""
Compute result of applying a chopper cascade to a neutron pulse at a time-of-flight
neutron source.

See :py:class:`FrameSequence` for the main entry point.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import numpy as np
import scipp as sc

from ..chopper import DiskChopper


def wavelength_to_inverse_velocity(wavelength):
    h = sc.constants.h
    m_n = sc.constants.m_n
    return (wavelength * m_n / h).to(unit='s/m')


def propagate_times(
    time: sc.Variable, wavelength: sc.Variable, distance: sc.Variable
) -> sc.Variable:
    """
    Propagate a neutron frame by a distance.

    Parameters
    ----------
    time:
        Time of the neutron frame.
    wavelength:
        Wavelength of the neutron frame.
    distance:
        Distance to propagate. Can be a range of distances.

    Returns
    -------
    :
        Propagated time.
    """
    inverse_velocity = wavelength_to_inverse_velocity(wavelength)
    return time + distance * inverse_velocity


class Subframe:
    """
    Neutron "subframe" at a time-of-flight neutron source, described as the corners of a
    polygon (initially a rectangle) in time and wavelength.
    """

    def __init__(self, time: sc.Variable, wavelength: sc.Variable):
        if {dim: time.sizes.get(dim) for dim in wavelength.sizes} != wavelength.sizes:
            raise sc.DimensionError(
                f'Inconsistent dims or shape: {time.sizes} vs {wavelength.sizes}'
            )
        self.time = time.to(unit='s', copy=False)
        self.wavelength = wavelength.to(unit='angstrom', copy=False)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Subframe):
            return NotImplemented
        return sc.identical(self.time, other.time) and sc.identical(
            self.wavelength, other.wavelength
        )

    def is_regular(self) -> bool:
        """
        Return True if the subframe is regular, i.e., if the vertex with the minimum
        wavelength also has the minimum time, and the vertex with the maximum wavelength
        also has the maximum time.
        """
        min_time = self.time == self.time.min()
        min_wavelength = self.wavelength == self.wavelength.min()
        max_time = self.time == self.time.max()
        max_wavelength = self.wavelength == self.wavelength.max()
        coinciding_min = min_time & min_wavelength
        coinciding_max = max_time & max_wavelength
        return coinciding_min.any() and coinciding_max.any()

    def propagate_by(self, distance: sc.Variable) -> Subframe:
        """
        Propagate subframe by a distance.

        Parameters
        ----------
        distance:
            Distance to propagate. Note that this is a difference, not an absolute
            value, in contrast to the distance in :py:meth:`Frame.propagate_to`.

        Returns
        -------
        :
            Propagated subframe.
        """
        return Subframe(
            time=propagate_times(self.time, self.wavelength, distance),
            wavelength=self.wavelength,
        )

    @property
    def start_time(self) -> sc.Variable:
        """The start time of the subframe for each of the distances in self.time."""
        # The `self.time` may have an additional dimension for distance, compared to
        # `self.wavelength`, and we need to keep that dimension in the output.
        out = self.time
        for dim in self.wavelength.dims:
            out = out.min(dim)
        return out

    @property
    def end_time(self) -> sc.Variable:
        """The end time of the subframe for each of the distances in self.time."""
        # The `self.time` may have an additional dimension for distance, compared to
        # `self.wavelength`, and we need to keep that dimension in the output.
        out = self.time
        for dim in self.wavelength.dims:
            out = out.max(dim)
        return out

    @property
    def start_wavelength(self) -> sc.Variable:
        """The start wavelength of the subframe for each of the distances in
        self.time"""
        return self.wavelength.min()

    @property
    def end_wavelength(self) -> sc.Variable:
        """The end wavelength of the subframe for each of the distances in self.time."""
        return self.wavelength.max()


@dataclass
class Frame:
    """
    A frame of neutrons, created from a single neutron pulse, potentially chopped into
    subframes by choppers.
    """

    distance: sc.Variable
    subframes: list[Subframe]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Frame):
            return NotImplemented
        # Note: we don't use sc.identical here because it can fail on the dtype even
        # if the values are equal.
        same_distance = (
            np.array_equal(self.distance.values, other.distance.values)
            and self.distance.unit == other.distance.unit
        )
        return same_distance and all(
            self_sub == other_sub
            for self_sub, other_sub in zip(self.subframes, other.subframes, strict=True)
        )

    def propagate_to(self, distance: sc.Variable) -> Frame:
        """
        Compute new frame by propagating to a distance.

        Parameters
        ----------
        distance:
            New distance.

        Returns
        -------
        :
            Propagated frame.
        """
        delta = distance - self.distance
        subframes = [subframe.propagate_by(delta) for subframe in self.subframes]
        return Frame(distance=distance, subframes=subframes)

    def chop(self, chopper: Chopper) -> Frame:
        """
        Compute a new frame by applying a chopper.

        A frame is a polygon in time and wavelength. Its initial shape is distorted
        by propagation to the chopper. The chopper then cuts off the parts of the frame
        that is outside of the chopper opening. Here we apply and algorithm that
        computes a new polygon that is the intersection of the frame and the chopper
        opening.

        In practice a chopper may have multiple openings, so a frame may be chopped into
        a number of subframes.

        Parameters
        ----------
        chopper:
            Chopper to apply.

        Returns
        -------
        :
            Chopped frame.
        """
        if chopper.distance < self.distance:
            raise ValueError(
                f'Chopper distance {chopper.distance} is smaller than frame distance '
                f'{self.distance}'
            )
        frame = self.propagate_to(chopper.distance)

        # A chopper can have multiple openings, call _chop for each of them. The result
        # is the union of the resulting subframes.
        chopped = Frame(distance=frame.distance, subframes=[])
        for subframe in frame.subframes:
            for open, close in zip(chopper.time_open, chopper.time_close, strict=True):
                if (tmp := _chop(subframe, open, close_to_open=True)) is not None:
                    if (tmp := _chop(tmp, close, close_to_open=False)) is not None:
                        chopped.subframes.append(tmp)
        return chopped

    def bounds(self) -> sc.DataGroup:
        """The bounds of the frame, i.e., the global min and max time and wavelength."""
        start = sc.reduce([sub.start_time for sub in self.subframes]).min()
        end = sc.reduce([sub.end_time for sub in self.subframes]).max()
        wav_start = sc.reduce([sub.start_wavelength for sub in self.subframes]).min()
        wav_end = sc.reduce([sub.end_wavelength for sub in self.subframes]).max()
        return sc.DataGroup(
            time=sc.concat([start, end], dim='bound'),
            wavelength=sc.concat([wav_start, wav_end], dim='bound'),
        )

    def subbounds(self) -> sc.DataGroup:
        """
        The bounds of the individual subframes, stored as a DataGroup.
        """
        starts = sc.concat(
            [subframe.start_time for subframe in self.subframes], dim='subframe'
        )
        ends = sc.concat(
            [subframe.end_time for subframe in self.subframes], dim='subframe'
        )
        # Given how time-propagation and chopping works, the min wavelength is always
        # given by the same vertex as the min time, and the max wavelength by the same
        # vertex as the max time. Thus, this check should generally always pass.
        # Exceptions may be subframes that have been created manually.
        if not all(subframe.is_regular() for subframe in self.subframes):
            raise NotImplementedError(
                'Subframes must be regular, i.e., min/max time and wavelength must '
                'coincide.'
            )
        wav_starts = sc.concat(
            [subframe.start_wavelength for subframe in self.subframes], dim='subframe'
        )
        wav_ends = sc.concat(
            [subframe.end_wavelength for subframe in self.subframes], dim='subframe'
        )

        time = sc.concat([starts, ends], dim='bound')
        wavelength = sc.concat([wav_starts, wav_ends], dim='bound')
        time_dims = list(set(time.dims) - {'subframe', 'bound'})
        wavelength_dims = list(set(wavelength.dims) - {'subframe', 'bound'})

        return sc.DataGroup(
            time=time.transpose([*time_dims, 'subframe', 'bound']),
            wavelength=wavelength.transpose([*wavelength_dims, 'subframe', 'bound']),
        )


@dataclass
class FrameSequence:
    """
    A sequence of frames, created from a single neutron pulse, potentially chopped into
    subframes by choppers.

    It is recommended to use the :py:meth:`from_source_pulse` constructor to create a
    frame sequence from a source pulse. Then, a chopper cascade can be applied using
    :py:meth:`chop`.
    """

    frames: list[Frame]

    @staticmethod
    def from_source_pulse(
        time_min: sc.Variable,
        time_max: sc.Variable,
        wavelength_min: sc.Variable,
        wavelength_max: sc.Variable,
    ):
        """
        Initialize a frame sequence from min/max time and wavelength of a pulse.

        The distance is set to 0 m.
        """
        time = sc.concat([time_min, time_max, time_max, time_min], dim='vertex').to(
            unit='s'
        )
        wavelength = sc.concat(
            [wavelength_min, wavelength_min, wavelength_max, wavelength_max],
            dim='vertex',
        ).to(unit='angstrom')
        frames = [
            Frame(
                distance=sc.scalar(0, unit='m'),
                subframes=[Subframe(time=time, wavelength=wavelength)],
            )
        ]
        return FrameSequence(frames)

    def __len__(self) -> int:
        """Number of frames."""
        return len(self.frames)

    def __getitem__(self, item: int | sc.Variable) -> Frame:
        """Get a frame by index or distance."""
        if isinstance(item, int):
            return self.frames[item]
        distance = item.to(unit='m')
        frame_before_detector = None
        for frame in self:
            if frame.distance > distance:
                break
            frame_before_detector = frame

        return frame_before_detector.propagate_to(distance)

    def propagate_to(self, distance: sc.Variable) -> FrameSequence:
        """
        Propagate the frame sequence to a distance, adding a new frame.

        Use this, e.g., to propagate to the sample position after applying choppers.

        Parameters
        ----------
        distance:
            Distance to propagate.

        Returns
        -------
        :
            New frame sequence.
        """
        return FrameSequence([*self.frames, self.frames[-1].propagate_to(distance)])

    def chop(self, choppers: list[Chopper]) -> FrameSequence:
        """
        Chop the frame sequence by a list of choppers.

        The choppers will be sorted by their distance, and applied in order.

        Parameters
        ----------
        choppers:
            List of choppers.

        Returns
        -------
        :
            New frame sequence.
        """
        choppers = sorted(choppers, key=lambda x: x.distance)
        frames = list(self.frames)
        for chopper in choppers:
            frames.append(frames[-1].chop(chopper))
        return FrameSequence(frames)

    def draw(
        self,
        linewidth: float = 0,
        fill: bool = True,
        alpha: float | None = None,
        transpose: bool = False,
        colors: list[str] | None = None,
        grid: bool = True,
        title: str = 'Frame propagation through chopper cascade',
        time_unit: str = 'ms',
        wavelength_unit: str = 'angstrom',
    ) -> Any:
        """
        Draw frames using matplotlib.

        Parameters
        ----------
        linewidth:
            Line width of frame edges.
        fill:
            Fill frame with color.
        alpha:
            Transparency of frame.
        transpose:
            Transpose axes.
        colors:
            List of colors to use for frames. If None, use default matplotlib colors.
        grid:
            Show grid.
        time_unit:
            Unit for time axis. Default is ms.
        wavelength_unit:
            Unit for wavelength axis. Default is angstrom.
        """
        import matplotlib.colors as mcolors
        import matplotlib.patches as patches
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        x_max = 0
        y_max = 0
        for i, frame in enumerate(self.frames):
            color = colors[i] if colors is not None else f'C{i}'
            # Add label to legend
            ax.plot([], [], color=color, label=f'{frame.distance:c}')
            # All subframes have same color
            for subframe in frame.subframes:
                x = subframe.time.to(unit=time_unit, copy=False)
                y = subframe.wavelength.to(unit=wavelength_unit, copy=False)
                if transpose:
                    x, y = y, x
                x_unit = x.unit
                y_unit = y.unit
                x_max = max(x_max, x.max().value)
                y_max = max(y_max, y.max().value)
                if alpha:
                    color = mcolors.to_rgba(color, alpha=alpha)
                polygon = patches.Polygon(
                    np.stack((x.values, y.values), axis=1),
                    closed=True,
                    fill=fill,
                    color=color,
                    linewidth=linewidth,
                )
                ax.add_patch(polygon)
        ax.set_xlabel(x_unit)
        ax.set_ylabel(y_unit)
        ax.set_xlim(0, x_max)
        ax.set_ylim(0, y_max)
        ax.minorticks_on()
        if grid:
            ax.grid(True, linestyle='-', linewidth='0.5', color='gray')
            ax.grid(which='minor', linestyle=':', linewidth='0.5', color='gray')
        ax.legend(loc='best')
        ax.set_title(title)
        fig.tight_layout()
        return fig, ax

    def acceptance_diagram(self):
        """
        Draw a chopper acceptance diagram.

        See, e.g., J.R.D. Copley, An acceptance diagram analysis of the contaminant
        pulse removal problem with direct geometry neutron chopper spectrometers,
        https://doi.org/10.1016/S0168-9002(03)01731-5 for more background.
        """
        import matplotlib.pyplot as plt

        source_distance = self.frames[0].distance
        frames = FrameSequence(
            [frame.propagate_to(source_distance) for frame in self.frames]
        )
        # Reset frame distance for plotting labels. This is a bit of a hack.
        # We should use chopper names if available.
        for i, frame in enumerate(frames.frames):
            frame.distance = self.frames[i].distance
        blue = np.linspace(0.1, 1, len(frames.frames))
        colors = plt.cm.Blues(blue)
        fig, ax = frames.draw(
            fill=True,
            linewidth=0.5,
            transpose=True,
            colors=colors,
            title='Chopper acceptance diagram',
        )
        # Put legend outside of plot
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(
            loc='center left',
            bbox_to_anchor=(1, 0.5),
            title='Distance',
            frameon=False,
        )
        return fig, ax


@dataclass
class Chopper:
    distance: sc.Variable
    time_open: sc.Variable
    time_close: sc.Variable

    def __post_init__(self):
        if self.time_open.sizes != self.time_close.sizes:
            raise sc.DimensionError(
                f'Inconsistent dims or shape: {self.time_open.sizes} vs '
                f'{self.time_close.sizes}'
            )

    def __getitem__(self, key) -> Chopper:
        return Chopper(
            distance=self.distance,
            time_open=self.time_open[key],
            time_close=self.time_close[key],
        )

    @classmethod
    def from_disk_chopper(
        cls,
        disk_chopper: DiskChopper,
        pulse_frequency: sc.Variable,
        npulses: int,
    ) -> Chopper:
        """
        Create a chopper from a NeXus disk chopper.

        Parameters
        ----------
        disk_chopper:
            Disk chopper.
        pulse_frequency:
            The frequency of pulses in the neutron source (this is to determine how
            many rotations the chopper performs per pulse).
        npulses:
            Number of pulses to rotate the chopper for.
        """
        tpulse = 1.0 / pulse_frequency
        topen = disk_chopper.time_offset_open(pulse_frequency=pulse_frequency)
        tclose = disk_chopper.time_offset_close(pulse_frequency=pulse_frequency)
        offsets = sc.arange('pulse', npulses) * tpulse
        return cls(
            distance=sc.norm(disk_chopper.axle_position),
            time_open=(offsets + topen).flatten(to=topen.dim),
            time_close=(offsets + tclose).flatten(to=tclose.dim),
        )


def _chop(frame: Subframe, time: sc.Variable, close_to_open: bool) -> Subframe | None:
    inside = frame.time >= time if close_to_open else frame.time <= time
    output = []
    for i in range(len(frame.time)):
        # Note how j wraps around to 0
        j = (i + 1) % len(frame.time)
        inside_i = inside[i]
        inside_j = inside[j]
        if inside_i:
            output.append((frame.time[i], frame.wavelength[i]))
        if inside_i != inside_j:
            # Intersection
            t = (time - frame.time[i]) / (frame.time[j] - frame.time[i])
            v = (1 - t) * frame.wavelength[i] + t * frame.wavelength[j]
            output.append((time, v))
    if not output:
        return None
    time = sc.concat([t for t, _ in output], dim=frame.time.dim)
    wavelength = sc.concat([v for _, v in output], dim=frame.wavelength.dim)
    return Subframe(time=time, wavelength=wavelength)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import string
from io import StringIO

import numpy as np
import pytest
import scipp as sc
from hypothesis import assume, given, settings
from hypothesis import strategies as st
from packaging.version import Version
from scipp.testing import strategies as scst

import scippneutron as scn


@st.composite
def one_dim_data_arrays(
    draw: st.DrawFn, min_n_coords: int = 1, max_n_coords: int = 5
) -> sc.DataArray:
    data = draw(scst.variables(ndim=1, dtype='float64', with_variances=True))
    # See https://github.com/scipp/scipp/issues/3052
    data.variances = abs(data.variances)
    coords = draw(
        st.dictionaries(
            keys=st.text(),
            values=scst.variables(
                sizes=data.sizes, dtype='float64', with_variances=False
            ),
            min_size=min_n_coords,
            max_size=max_n_coords,
        )
    )
    return sc.DataArray(data, coords=coords)


def headers() -> st.SearchStrategy[str]:
    # Using only ASCII characters and excluding \r because Numpy and/or splitlines
    # do something funny with some other characters.
    return st.text(string.ascii_letters + string.digits + string.punctuation + ' \t\n')


def save_to_buffer(da: sc.DataArray, coord: str | None = None, **kwargs) -> StringIO:
    buffer = StringIO()
    scn.io.save_xye(buffer, da, coord=coord, **kwargs)
    buffer.seek(0)
    return buffer


def roundtrip(da: sc.DataArray, coord: str | None = None, **kwargs) -> sc.DataArray:
    buffer = save_to_buffer(da, coord, **kwargs)
    return scn.io.xye.load_xye(
        buffer,
        dim=da.dim,
        coord=coord,
        unit=da.unit,
        coord_unit=da.coords[coord].unit if coord is not None else None,
    )


@given(initial=one_dim_data_arrays(), header=headers(), data=st.data())
def test_roundtrip(initial, header, data):
    coord_name = data.draw(st.sampled_from(list(initial.coords.keys())))
    loaded = roundtrip(initial, coord=coord_name, header=header)
    assert set(loaded.coords.keys()) == {coord_name}
    # Using allclose instead of identical because the format might lose some precision.
    # Especially in the variances -> stddevs conversion.
    assert sc.allclose(
        loaded.coords[coord_name], initial.coords[coord_name], equal_nan=True
    )
    assert sc.allclose(loaded.data, initial.data, equal_nan=True)


@given(da=one_dim_data_arrays(), data=st.data())
def test_saved_file_contains_data_table(da, data):
    coord_name = data.draw(st.sampled_from(list(da.coords.keys())))
    file_contents = save_to_buffer(da, coord=coord_name).getvalue()
    for i, line in enumerate(
        filter(
            lambda line: line and not line.startswith('#'), file_contents.split('\n')
        )
    ):
        x, y, e = map(float, line.split(' '))
        np.testing.assert_allclose(x, da.coords[coord_name][i].value)
        np.testing.assert_allclose(y, da[i].value)
        np.testing.assert_allclose(e, np.sqrt(da[i].variance))


@given(initial=one_dim_data_arrays(max_n_coords=1))
@settings(max_examples=20)
def test_save_deduce_coord_data_with_one_coord(initial):
    coord_name = next(iter(initial.coords.keys()))
    loaded = roundtrip(initial)
    loaded = loaded.rename_dims({loaded.dim: initial.dim})
    # roundtrip cannot deduce coord name and unit in this case.
    loaded_coord = next(iter(loaded.coords.values()))
    loaded_coord.unit = initial.coords[coord_name].unit
    assert sc.allclose(loaded_coord, initial.coords[coord_name], equal_nan=True)


@given(initial=one_dim_data_arrays(), data=st.data())
@settings(max_examples=20)
def test_save_deduce_coord_dim_coord(initial, data):
    dim = initial.dim
    initial_coord = data.draw(
        scst.variables(sizes=initial.sizes, dtype='float64', with_variances=False)
    )
    initial.coords[dim] = initial_coord.rename({initial_coord.dim: dim})

    loaded = roundtrip(initial)
    # roundtrip cannot deduce coord name and unit in this case.
    # loaded_coord = next(iter(loaded.coords.values()))
    loaded_coord = loaded.coords[dim]
    loaded_coord.unit = initial.coords[dim].unit
    assert sc.allclose(loaded_coord, initial.coords[dim], equal_nan=True)


@given(da=one_dim_data_arrays(min_n_coords=2))
@settings(max_examples=20)
def test_save_cannot_deduce_coord_if_multiple_non_dim_coords(da):
    # It can deduce if there is a dim-coord, exclude that.
    assume(da.dim not in da.coords)
    with pytest.raises(
        ValueError,
        match='Cannot deduce which coordinate to save because the data '
        'has more than one and no dimension-coordinate',
    ):
        save_to_buffer(da)


@given(da=one_dim_data_arrays(max_n_coords=1))
@settings(max_examples=20)
def test_input_must_have_at_least_one_coord(da):
    for c in list(da.coords.keys()):
        del da.coords[c]
    with pytest.raises(
        ValueError, match='Cannot save data to XYE file because it has no coordinates'
    ):
        save_to_buffer(da)


@given(da=one_dim_data_arrays(), data=st.data())
@settings(max_examples=20)
def test_input_must_have_variances(da, data):
    da.variances = None
    coord_name = data.draw(st.sampled_from(list(da.coords.keys())))
    with pytest.raises(sc.VariancesError):
        save_to_buffer(da, coord=coord_name)


@given(da=one_dim_data_arrays(), data=st.data())
@settings(max_examples=20)
def test_cannot_save_data_with_bin_edges(da, data):
    coord_name = data.draw(st.sampled_from(list(da.coords.keys())))
    da.coords[coord_name] = sc.concat(
        [da.coords[coord_name], sc.scalar(0.0, unit=da.coords[coord_name].unit)],
        dim=da.dim,
    )
    with pytest.raises(sc.CoordError):
        save_to_buffer(da, coord=coord_name)


@given(da=one_dim_data_arrays(), data=st.data())
@settings(max_examples=20)
def test_cannot_save_data_with_masks(da, data):
    mask = data.draw(scst.variables(sizes=da.sizes, dtype=bool))
    da.masks[data.draw(st.text())] = mask
    with pytest.raises(
        ValueError, match='Cannot save data to XYE file because it has masks'
    ):
        save_to_buffer(da, coord=next(iter(da.coords)))


@pytest.mark.skipif(
    Version(sc.__version__) < Version('24.1'),
    reason='This use of the dataarrays strategy needs Scipp >= 24.*',
)
@given(data=st.data())
@settings(max_examples=20)
def test_input_must_be_one_dimensional(data):
    da = data.draw(
        scst.dataarrays(
            data_args={
                'ndim': st.integers(min_value=2, max_value=4),
                'dtype': 'float64',
                'with_variances': True,
            },
            masks=False,
            bin_edges=False,
        )
    )
    assume(da.coords)
    coord_name = data.draw(st.sampled_from(list(da.coords.keys())))
    with pytest.raises(sc.DimensionError):
        save_to_buffer(da, coord=coord_name)


@given(da=one_dim_data_arrays(), header=headers())
def test_can_set_header(da, header):
    buffer = save_to_buffer(da, coord=next(iter(da.coords)), header=header)
    commented_header = (
        '\n'.join(f'# {line}' for line in header.splitlines()) if header else ''
    )
    assert buffer.getvalue().startswith(commented_header)


@given(
    da=one_dim_data_arrays(), coord_name=st.text(string.ascii_letters + string.digits)
)
def test_generated_header_includes_coord_name_and_units(da, coord_name):
    # Detecting the coord name in the file can be tricky if it can contain arbitrary
    # characters (like '\n' or '#') because lines in the header start with '# '.
    initial_name = next(iter(da.coords))
    da.coords[coord_name] = da.coords.pop(initial_name)

    buffer = save_to_buffer(da, coord=coord_name)
    assert coord_name in buffer.getvalue()
    if da.coords[coord_name].unit is not None:
        assert str(da.coords[coord_name].unit) in buffer.getvalue()
    if da.unit is not None:
        assert str(da.unit) in buffer.getvalue()


def test_loads_correct_values():
    file_contents = '''1 2 3
1.003 32.1 5
0.1111 0 2.1e-3
'''
    buffer = StringIO(file_contents)
    loaded = scn.io.load_xye(buffer, dim='my-dim', unit='one', coord_unit='us')
    expected = sc.DataArray(
        sc.array(
            dims=['my-dim'],
            values=[2, 32.1, 0],
            variances=np.power([3, 5, 2.1e-3], 2),
            unit='one',
        ),
        coords={
            'my-dim': sc.array(dims=['my-dim'], values=[1, 1.003, 0.1111], unit='us')
        },
    )
    assert sc.identical(loaded, expected)



import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_allclose

from scippneutron.absorption.cylinder import Cylinder


@pytest.fixture(
    params=[
        sc.vector([-0.9854498123542775, -0.1699666653521188, 0]),
        sc.vector([0.7242293076582138, 0.6895592142295717, 0]),
        sc.vector([0.9445601257691059, 0.3283384972966938, 0]),
        sc.vector([-0.8107703958381735, -0.5853643012965615, 0]),
        sc.vector([0.48126033718549477, -0.8765777135269319, 0]),
    ]
)
def point_on_unit_circle(request):
    return request.param


@pytest.fixture(params=[sc.scalar(1.2), sc.scalar(0.2)])
def height(request):
    return request.param


@pytest.fixture(params=[sc.scalar(1.2), sc.scalar(0.2)])
def radius(request):
    return request.param


@pytest.fixture(
    params=[
        sc.vector([0.0, 0.0, 1.0]),
        sc.vector([-0.5620126808026259, -0.1798933776079791, 0.8073290031392648]),
    ]
)
def axis(request):
    return request.param


@pytest.fixture(
    params=[
        sc.vector([0.0, 0.0, 0.0]),
        sc.vector([1.0, -2.0, 3.0]),
    ]
)
def base(request):
    return request.param


@pytest.fixture
def cylinder(request, axis, base, radius, height):
    return Cylinder(axis, base, radius, height)


def _rotate_from_z_to_axis(p, ax):
    z = sc.vector([0, 0, 1.0])
    if ax == z:
        return p
    u = sc.cross(z, ax)
    un = sc.norm(u)
    u *= sc.asin(un) / un
    return sc.spatial.rotations_from_rotvecs(u) * p


def test_intersection_in_base(cylinder, point_on_unit_circle):
    v = _rotate_from_z_to_axis(point_on_unit_circle, cylinder.symmetry_line)
    assert_allclose(
        cylinder.beam_intersection(
            cylinder.radius * v
            + cylinder.center_of_base
            # Move start point just inside cylinder.
            # Required to make all tests pass.
            + 2 * np.finfo(float).eps * cylinder.symmetry_line,
            -v,
        ),
        2 * cylinder.radius,
    )


def test_intersection_diagonal(cylinder, point_on_unit_circle):
    v = _rotate_from_z_to_axis(point_on_unit_circle, cylinder.symmetry_line)
    n = -2 * cylinder.radius * v + cylinder.symmetry_line * cylinder.height
    n /= sc.norm(n)
    # Diagonal intersection is as expected
    assert_allclose(
        cylinder.beam_intersection(cylinder.radius * v + cylinder.center_of_base, n),
        ((2 * cylinder.radius) ** 2 + cylinder.height**2) ** 0.5,
    )
    # Intersection is zero in other directions
    assert_allclose(
        cylinder.beam_intersection(cylinder.radius * v + cylinder.center_of_base, -n),
        sc.scalar(0.0),
        atol=sc.scalar(1e-14),
    )
    n = 2 * cylinder.radius * v + cylinder.symmetry_line * cylinder.height
    n /= sc.norm(n)
    assert_allclose(
        cylinder.beam_intersection(cylinder.radius * v + cylinder.center_of_base, n),
        sc.scalar(0.0),
        atol=sc.scalar(1e-14),
    )
    assert_allclose(
        cylinder.beam_intersection(cylinder.radius * v + cylinder.center_of_base, -n),
        sc.scalar(0.0),
        atol=sc.scalar(1e-14),
    )


def test_intersection_along_axis(cylinder, point_on_unit_circle):
    v = _rotate_from_z_to_axis(point_on_unit_circle, cylinder.symmetry_line)
    # Move point just inside cylinder
    x = cylinder.radius * v + cylinder.center_of_base - 2 * np.finfo(float).eps * v
    assert_allclose(
        cylinder.beam_intersection(x, cylinder.symmetry_line), cylinder.height
    )


def test_no_intersection(cylinder, point_on_unit_circle):
    v = _rotate_from_z_to_axis(point_on_unit_circle, cylinder.symmetry_line)
    x = cylinder.center_of_base + cylinder.radius * v
    assert_allclose(
        cylinder.beam_intersection(x, v), sc.scalar(0.0), atol=sc.scalar(1e-14)
    )
    # Rotate 90 deg around cylinder axis.
    # Should still not intersect the cylinder.
    v = (
        sc.spatial.rotations_from_rotvecs(
            cylinder.symmetry_line * sc.scalar(np.pi / 2, unit='rad')
        )
        * v
    )
    assert_allclose(
        cylinder.beam_intersection(x, v), sc.scalar(0.0), atol=sc.scalar(1e-7)
    )


def test_intersection_from_center(cylinder, point_on_unit_circle):
    v = _rotate_from_z_to_axis(point_on_unit_circle, cylinder.symmetry_line)
    assert_allclose(cylinder.beam_intersection(cylinder.center, v), cylinder.radius)
    assert_allclose(
        cylinder.beam_intersection(cylinder.center, cylinder.symmetry_line),
        cylinder.height / 2,
    )
    assert_allclose(
        cylinder.beam_intersection(cylinder.center, -cylinder.symmetry_line),
        cylinder.height / 2,
    )


@pytest.mark.parametrize('kind', ['expensive', 'medium', 'cheap'])
@pytest.mark.parametrize(
    ('f', 'expected'),
    [
        ((lambda x: 1.0), sc.scalar(np.pi)),
        ((lambda x: sc.dot(x, sc.vector([0, 0, 1]))), sc.scalar(np.pi / 2)),
        ((lambda x: sc.dot(x, sc.vector([0, 1, 0]))), sc.scalar(0.0)),
        ((lambda x: sc.dot(x, sc.vector([1, 0, 0]))), sc.scalar(0.0)),
        ((lambda x: x.fields.x * x.fields.y), sc.scalar(0.0)),
        (
            (
                lambda x: sc.sin(
                    sc.dot(x, sc.vector([1, 0.5, 0]) * sc.scalar(1, unit='rad'))
                )
                ** 2
            ),
            # Exact value from Sympy
            sc.scalar(0.7972445081889596),
        ),
    ],
)
def test_quadrature(kind, f, expected):
    c = Cylinder(
        sc.vector([0, 0, 1.0]), sc.vector([0, 0, 0.0]), sc.scalar(1.0), sc.scalar(1.0)
    )
    q, w = c.quadrature(kind)
    v = (f(q) * w).sum()
    assert_allclose(v, expected, atol=sc.scalar(1e-8), rtol=sc.scalar(1e-4))


@pytest.mark.parametrize('kind', ['mc', ('mc', 4000)])
def test_quadrature_mc(kind):
    import numpy

    numpy.random.seed(1)

    c = Cylinder(
        sc.vector([0, 0, 1.0]), sc.vector([0, 0, 0.0]), sc.scalar(1.0), sc.scalar(1.0)
    )

    def f(x):
        return sc.sin(sc.dot(x, sc.vector([1, 0.5, 0]) * sc.scalar(1, unit='rad'))) ** 2

    expected = sc.scalar(0.7972445081889596)
    q, w = c.quadrature(kind)
    v = (f(q) * w).sum()
    assert_allclose(v, expected, atol=sc.scalar(5e-2), rtol=sc.scalar(5e-2))

    if isinstance(kind, tuple):
        assert len(q) == kind[1]


@pytest.mark.parametrize('kind', ['expensive', 'medium', 'cheap'])
@pytest.mark.parametrize(
    'axis', [sc.vector([0, 1, 0.0]), sc.vector([2**-0.5, 0, 2**-0.5])]
)
@pytest.mark.parametrize('base', [sc.vector([0, 0, 0.0]), sc.vector([2.0, 5.3, -4.0])])
@pytest.mark.parametrize('r', [sc.scalar(0.2), sc.scalar(1.0)])
@pytest.mark.parametrize('h', [sc.scalar(0.2), sc.scalar(1.0)])
def test_quadrature_translated_cylinder(kind, axis, base, r, h):
    c = Cylinder(axis, base, r, h)
    q, w = c.quadrature(kind)

    def f(x):
        return sc.dot(axis / h, x - base)

    v = (f(q) * w).sum()
    assert_allclose(v, r**2 * h * np.pi / 2, atol=sc.scalar(1e-12))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
"""
This module provides functionality for unwrapping raw frames of neutron time-of-flight
data.

The module handles standard unwrapping, unwrapping in pulse-skipping mode, and
unwrapping for WFM instruments, as well as combinations of the latter two. The
functions defined here are meant to be used as providers for a Sciline pipeline. See
https://scipp.github.io/sciline/ on how to use Sciline.
"""

import math
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import NewType

import numpy as np
import scipp as sc

from .._utils import elem_unit
from . import chopper_cascade

ChopperCascadeFrames = NewType('ChopperCascadeFrames', chopper_cascade.FrameSequence)
"""
Result of passing the source pulse through the chopper cascade.
"""


Choppers = NewType('Choppers', Mapping[str, chopper_cascade.Chopper])
"""
Choppers used to define the frame parameters.
"""


FrameAtSample = NewType('FrameAtSample', chopper_cascade.Frame)
"""
Result of passing the source pulse through the chopper cascade.

After propagating the frame to the detector distances this will then be used to
compute the frame bounds.
"""

FrameAtDetector = NewType('FrameAtDetector', chopper_cascade.Frame)
"""
Result of passing the source pulse through the chopper cascade to the detector.

The detector may be a monitor or a detector after scattering off the sample. The frame
bounds are then computed from this.
"""

CleanFrameAtDetector = NewType('CleanFrameAtDetector', chopper_cascade.Frame)
"""
Version of the frame at the detector with subframes that do not overlap.
"""

FrameBounds = NewType('FrameBounds', sc.DataGroup)
"""
The computed frame boundaries, used to unwrap the raw timestamps.
"""

FrameForTimeOfFlightOrigin = NewType(
    'FrameForTimeOfFlightOrigin', chopper_cascade.Frame
)
"""
Frame used to compute the time-of-flight origin. Used in WFM.
"""

FramePeriod = NewType('FramePeriod', sc.Variable)
"""
The period of a frame, a (small) integer multiple of the source period.
"""

FrameWrappedTimeOffset = NewType('FrameWrappedTimeOffset', sc.Variable)
"""
Time offsets wrapped around at the frame period.

In normal mode, this is identical to the NXevent_data/event_time_offset recorded by
the data acquisition system, i.e., the same as :py:class:`PulseWrappedTimeOffset`.

In pulse-skipping mode, this is the time offset since the start of the frame.
"""

Ltotal = NewType('Ltotal', sc.Variable)
"""
Total distance between the source and the detector(s).

This is used to propagate the frame to the detector position. This will then yield
detector-dependent frame bounds. This is typically the sum of L1 and L2, except for
monitors.
"""

PulseOffset = NewType('PulseOffset', sc.Variable)
"""
Offset from the start of the frame in pulse-skipping mode, multiple of pulse period.
"""

PulsePeriod = NewType('PulsePeriod', sc.Variable)
"""
Period of the source pulses, i.e., time between consecutive pulse starts.
"""

PulseStride = NewType('PulseStride', int)
"""
Stride of used pulses. Usually 1, but may be a small integer when pulse-skipping.
"""

PulseWrappedTimeOffset = NewType('PulseWrappedTimeOffset', sc.Variable)
"""
Time offsets wrapped at the pulse period, typically NXevent_data/event_time_offset.
"""

RawData = NewType('RawData', sc.DataArray)
"""
Raw detector data loaded from a NeXus file, e.g., NXdetector containing NXevent_data.
"""

SourceChopperName = NewType('SourceChopperName', str)
"""
Name of the chopper defining the source location and time-of-flight time origin.
"""

SourceChopper = NewType('SourceChopper', chopper_cascade.Chopper)
"""
Chopper defining the source location and time-of-flight time origin.

If there is no pulse-shaping chopper, a fake chopper with time_open and time_close
set to the source pulse begin and end time should be used.
"""

SampleDistance = NewType('SampleDistance', sc.Variable)
"""
Location of the sample along the incident beam. Origin must be consistent with chopper
distance origin.
"""

SourceTimeRange = NewType('SourceTimeRange', tuple[sc.Variable, sc.Variable])
"""
Time range of the source pulse, used for computing frame bounds.
"""

SourceWavelengthRange = NewType(
    'SourceWavelengthRange', tuple[sc.Variable, sc.Variable]
)
"""
Wavelength range of the source pulse, used for computing frame bounds.
"""

TimeZero = NewType('TimeZero', sc.Variable)
"""
Time of the start of the most recent pulse, typically NXevent_data/event_time_zero.
"""

UnwrappedData = NewType('UnwrappedData', sc.DataArray)
"""
Detector data with unwrapped time offset and pulse time coordinates.
"""

TimeOfFlightOriginDistance = NewType('TimeOfFlightOriginDistance', sc.Variable)
"""
Distance from the source to the position the time-of-flight origin.
"""

TimeOfFlightOriginTime = NewType('TimeOfFlightOriginTime', sc.Variable)
"""
Time of the time-of-flight origin for each subframe.
"""

TofData = NewType('TofData', sc.DataArray)
"""
Detector data with time-of-flight and time zero coordinates.
"""


TimeOffset = NewType('TimeOffset', sc.Variable)
"""Unwrapped time offset relative to the pulse time."""

DeltaFromWrapped = NewType('DeltaFromWrapped', sc.Variable)
"""Positive delta to be added to the input time offsets to unwrap them."""


@dataclass
class TimeOfFlightOrigin:
    """
    The origin of the time-of-flight, time since pulse time and distance from source.
    """

    time: sc.Variable | sc.DataArray
    distance: sc.Variable


def frame_period(
    pulse_period: PulsePeriod, pulse_stride: PulseStride | None
) -> FramePeriod:
    if pulse_stride is None:
        return pulse_period
    return FramePeriod(pulse_period * pulse_stride)


def chopper_cascade_frames(
    source_wavelength_range: SourceWavelengthRange,
    source_time_range: SourceTimeRange,
    choppers: Choppers,
) -> ChopperCascadeFrames:
    """
    Return all the chopper frames.

    This is the result of propagating the source pulse through the chopper cascade.

    It is assumed that the opening and closing times of the input choppers have been
    setup correctly. This includes a correct definition of the offsets in
    pulse-skipping mode, i.e., the caller must know which pulses are in use.
    """
    frames = chopper_cascade.FrameSequence.from_source_pulse(
        time_min=source_time_range[0],
        time_max=source_time_range[-1],
        wavelength_min=source_wavelength_range[0],
        wavelength_max=source_wavelength_range[-1],
    )
    return ChopperCascadeFrames(frames.chop(choppers.values()))


def frame_at_detector(
    frames: ChopperCascadeFrames,
    ltotal: Ltotal,
) -> FrameAtDetector:
    """
    Return the frame at the detector.

    This is the result of propagating the source pulse through the chopper cascade to
    the detector. The detector may be a monitor or a detector after scattering off the
    sample. The frame bounds are then computed from this.
    """
    return FrameAtDetector(frames[-1].propagate_to(ltotal))


def frame_bounds(frame: FrameAtDetector) -> FrameBounds:
    return FrameBounds(frame.bounds())


def frame_wrapped_time_offset(offset: PulseWrappedTimeOffset) -> FrameWrappedTimeOffset:
    """Without pulse-skipping, this is an identity function."""
    return FrameWrappedTimeOffset(offset)


def pulse_offset(
    pulse_period: PulsePeriod,
    pulse_stride: PulseStride | None,
    event_time_zero: TimeZero,
) -> PulseOffset:
    """
    Return the offset of a pulse within a frame.

    This has no special meaning, since on its own it does not define which pulse
    is used for the frame. Instead of taking this into account here, the SourceChopper
    used to compute OffsetFromTimeOfFlight will take care of this. This assumes that
    the choppers have been defined correctly.
    """
    if pulse_stride is None:
        return PulseOffset(sc.scalar(0, unit=elem_unit(pulse_period)))
    first_pulse_time = (
        event_time_zero[0]
        if len(event_time_zero) > 0
        else sc.zeros_like(event_time_zero)
    )
    # This is roughly equivalent to
    #   (event_time_zero - first_pulse_time) % frame_period
    # but should avoid some issues with precision and drift
    pulse_index = (event_time_zero - first_pulse_time) // pulse_period.to(
        unit=elem_unit(event_time_zero)
    )
    return PulseOffset(pulse_period * (pulse_index % pulse_stride))


def frame_wrapped_time_offset_pulse_skipping(
    offset: PulseWrappedTimeOffset, pulse_offset: PulseOffset
) -> FrameWrappedTimeOffset:
    """Return the time offset wrapped around at the frame period."""
    return FrameWrappedTimeOffset(offset + pulse_offset.to(unit=elem_unit(offset)))


def pulse_wrapped_time_offset(da: RawData) -> PulseWrappedTimeOffset:
    """
    In NXevent_data, event_time_offset is the time offset since event_time_zero,
    which is the start of the most recent pulse. This is not the same as the start
    of the pulse that emitted the neutron, so event_time_offset is "wrapped" around
    at the pulse period.
    """
    if da.bins is None:
        # Canonical name in NXmonitor
        return PulseWrappedTimeOffset(da.coords['time_of_flight'])
    return PulseWrappedTimeOffset(da.bins.coords['event_time_offset'])


def time_zero(da: RawData) -> TimeZero:
    """
    In NXevent_data, event_time_zero is the start of the most recent pulse.
    """
    # This is not available for histogram-mode monitors. We probably need to look it
    # up elsewhere in the file, e.g., NXsource. Should we have a separate provider for
    # monitors?
    if da.bins is None:
        raise NotImplementedError(
            "NXevent_data/event_time_zero not available for histogram-mode monitors."
        )
    return da.coords['event_time_zero']


def offset_from_wrapped(
    wrapped_time_offset: FrameWrappedTimeOffset,
    frame_bounds: FrameBounds,
    frame_period: FramePeriod,
) -> DeltaFromWrapped:
    """
    Offset of the input time offsets from the start of the frame emitting the neutron.

    In other words, this is the part that is "lost" by the conceptual wrapping of the
    time offsets at the frame period. This is the offset that needs to be added to the
    NXevent_data/event_time_offset to obtain the offset from the start of the emitting
    frame.

    This is not identical to the offset to time-of-flight, since the time-of-flight is
    measured, e.g., from the center of the pulse, to the center of a pulse-shaping
    chopper slit.

    Note that this takes the frame-period, which can be a multiple of a pulse-period.
    The input time stamps must be relative the the start of the frame, i.e., in
    pulse-skipping mode a pulse offset must have been applied before calling this
    function.

    Parameters
    ----------
    wrapped_time_offset :
        Time offset from the time-zero as recorded by the data acquisition system.
    frame_bounds :
        The computed frame boundaries, used to unwrap the raw timestamps.
        Typically pixel-dependent when unwrapping detectors.
    frame_period :
        Time between the start of two consecutive frames, i.e., the period of the
        time-zero used by the data acquisition system.
    """
    time_bounds = frame_bounds['time']
    frame_period = frame_period.to(unit=elem_unit(time_bounds))
    diff = (time_bounds['bound', -1] - time_bounds['bound', 0]) - frame_period
    if (diff > sc.scalar(0.0, unit=frame_period.unit)).any():
        raise ValueError(
            "Frames are overlapping: Computed frame bounds "
            f"{frame_bounds} are larger than frame period {frame_period}."
        )
    time_offset_min = time_bounds['bound', 0]
    wrapped_time_min = time_offset_min % frame_period
    # We simply cut, without special handling of times that fall outside the frame
    # period. Everything below and above is allowed. The alternative would be to, e.g.,
    # replace invalid inputs with NaN, but this would probably cause more trouble down
    # the line.
    begin = sc.full_like(wrapped_time_min, value=-math.inf)
    end = sc.full_like(wrapped_time_min, value=math.inf)
    dim = 'section'
    time = sc.concat([begin, wrapped_time_min, end], dim).transpose().copy()
    offset = sc.DataArray(
        time_offset_min
        - wrapped_time_min
        + sc.concat([frame_period, sc.zeros_like(frame_period)], dim),
        coords={dim: time.to(unit=elem_unit(wrapped_time_offset))},
    )
    return DeltaFromWrapped(sc.lookup(offset, dim=dim)[wrapped_time_offset])


def source_chopper(
    choppers: Choppers,
    source_time_range: SourceTimeRange,
    source_chopper_name: SourceChopperName | None,
) -> SourceChopper:
    """
    Return the chopper defining the source location and time-of-flight time origin.

    If there is no pulse-shaping chopper, then the source-pulse begin and end time
    are used to define a fake chopper.
    """
    if source_chopper_name is not None:
        return choppers[source_chopper_name]
    return chopper_cascade.Chopper(
        distance=sc.scalar(0.0, unit='m'),
        time_open=sc.concat([source_time_range[0]], 'cutout').to(unit='s'),
        time_close=sc.concat([source_time_range[1]], 'cutout').to(unit='s'),
    )


def time_of_flight_origin_from_chopper(
    source_chopper: SourceChopper,
) -> TimeOfFlightOrigin:
    """
    Compute the time-of-flight origin from a source chopper.

    This is a naive approach for defining the time of flight. We are not sure this
    is going to be used in practice. An possible alternative might be to calibrate
    the time-of-flight using a sample with a known Bragg edge. This is not implemented
    currently.

    A chopper is used to define (1) the "source" location and (2) the time-of-flight
    time origin. The time-of-flight is then the time difference between the time of
    arrival of the neutron at the detector, and the time of arrival of the neutron at
    the chopper. L1 needs to be redefined to be the distance between the chopper and
    the sample.

    If there is no pulse-shaping chopper, then the source-pulse begin and end time
    should be set as the source_time_open and source_time_close, respectively.

    For WFM, the source_time_open and source_time_close will be different for each
    subframe. In this case, all input parameters should be given as variables with
    subframe as dimension.

    Parameters
    ----------
    source_chopper :
        Chopper defining the source location and time-of-flight time origin (as the
        center of the slit opening). The chopper distance is assumed to be the distance
        from the source position.
    """
    if len(source_chopper.time_open) != 1:
        raise NotImplementedError(
            "Source chopper with multiple openings not supported yet."
        )
    source_time_open = source_chopper.time_open[0]
    source_time_close = source_chopper.time_close[0]
    time_zero = 0.5 * (source_time_open + source_time_close)
    return TimeOfFlightOrigin(time=time_zero, distance=source_chopper.distance)


def time_offset(unwrapped: UnwrappedData) -> TimeOffset:
    """
    Extract the time offset coord of the unwrapped input data.
    """
    if unwrapped.bins is not None:
        return TimeOffset(unwrapped.bins.coords['time_offset'])
    return TimeOffset(unwrapped.coords['time_offset'])


def maybe_clip_detector_subframes(
    frame: FrameAtDetector, ltotal: Ltotal, chopper_cascade_frames: ChopperCascadeFrames
) -> CleanFrameAtDetector:
    """
    Check for time overlap between subframes.
    If overlap is found, we clip away the regions where there is overlap.
    This is done by adding a fake chopper which is closed during the overlapping times.

    Examples:

    .. code-block:: text

       1. partial overlap:
       |-------------|            ->   |--------|
                |-------------|   ->                 |--------|

       2. total overlap:
       |----------------------|   ->   |--------|       |-----|
                |-------|         ->

    In the case of multiple detector pixels, we find the pixel closest to the source
    and use that as the reference for the clipping.
    Because we are removing the entire region where overlap occurs, we are inserting
    a gap between subframes. The gap starts at the start of subframe n+1 and ends at
    the end of subframe n.
    The distance it would take for the gap to be entirely filled again with overlap
    is usually greater than the range of pixel distances (some edge cases are possible).

    TODO: We currently do not handle the case where there is overlap between subframes
    at the detector farthest from the source, and no overlap at the detector closest to
    the source. Although rare, this is not impossible.

    Parameters
    ----------
    frame:
        The frame at the detector, with subframes that may overlap.
    ltotal:
        The total distance between the source and the detector.
    chopper_cascade_frames:
        All the frames in the chopper cascade, before the detector.
    """
    # Need subframe dim at the end because sorting requires contiguous data
    dims = (*ltotal.dims, 'subframe')
    starts = sc.concat(
        [sf.start_time for sf in frame.subframes], dim='subframe'
    ).transpose(dims)
    ends = sc.concat([sf.end_time for sf in frame.subframes], dim='subframe').transpose(
        dims
    )

    sizes = starts.sizes
    sizes['subframe'] *= 2
    bounds = sc.empty(sizes=sizes, unit=starts.unit)
    bounds['subframe', ::2] = sc.sort(starts, 'subframe')
    bounds['subframe', 1::2] = sc.sort(ends, 'subframe')
    if all(sc.issorted(bounds, 'subframe').flatten(to='x')):
        return CleanFrameAtDetector(frame)

    sorted_times = sc.sort(bounds, 'subframe')
    if ltotal.dims:
        # Find the pixel closest to the source
        closest_pixel = np.argmin(ltotal.flatten(to='pixel').values)
        sorted_times = sorted_times.flatten(dims=ltotal.dims, to='pixel')[
            'pixel', closest_pixel
        ]

    # Make a fake chopper that is closed during the overlapping times, placed
    # immediately before the closest detector pixel.
    fake_chopper = chopper_cascade.Chopper(
        distance=frame.distance.min(),
        time_open=sorted_times['subframe', ::2],
        time_close=sorted_times['subframe', 1::2],
    )
    # We cannot chop frames at multiple distances at once, so instead of chopping the
    # frame at the detector, we chop the last frame in the chopper cascade before the
    # detector.
    last_frame = chopper_cascade_frames[-1]

    # Chop the subframes one by one
    # TODO: We currently need to chop them one by one as the `chop` method seems
    # to not work correctly with overlapping subframes.
    subframes = []
    for subframe in last_frame.subframes:
        f = chopper_cascade.Frame(distance=last_frame.distance, subframes=[subframe])
        chopped = f.chop(fake_chopper)
        subframes.extend(
            [
                sf
                for sf in chopped.subframes
                if not sc.allclose(sf.start_time, sf.end_time)
            ]
        )

    return CleanFrameAtDetector(
        chopper_cascade.Frame(
            distance=fake_chopper.distance,
            subframes=subframes,
        ).propagate_to(frame.distance)
    )


def time_of_flight_origin_wfm(
    frames: ChopperCascadeFrames,
    clean_detector_frame: CleanFrameAtDetector,
) -> TimeOfFlightOrigin:
    """
    Compute the time-of-flight origin in the WFM case.
    For WFM there is not a single time-of-flight "origin", but one for each subframe.

    Parameters
    ----------
    frames:
        All the frames in the chopper cascade.
    clean_detector_frame:
        The frame at the detector, with subframes that do not overlap.

    Notes
    -----
    To find the time-of-flight origin, we ray-trace the fastest and slowest neutrons of
    the subframes back to the first chopper to determine the time-of-flight origin.
    The assumption here is that the first chopper is one of the two wfm choppers.
    We also assume that the first chopper will be inundated with neutrons from the
    source, and propagating the boundaries of the frame backwards should thus give
    us a good estimate of the opening and closing times of the first chopper. There is
    also the general rule that the longer the instrument, the better the wavelength
    resolution, so going back to the first chopper also makes sense.

    We use this method instead of reading the opening and closing times of the choppers
    because it is not possible to know in a trivial way which opening and closing times
    correspond to which subframe at the detector. There is not always a 1-1
    mapping between the subframes at the detector and the chopper cutouts (for instance,
    the DREAM pulse shaping choppers have 8 cutouts, but typically create 2 subframes
    at the detector). In addition, if the chopper is rotating rapidly, there may be
    multiple opening and closing times, where the extra subframes get blocked by other
    choppers further down the beamline.

    While developing this method, we attempted several other implementations, which all
    gave worse results than the current implementation. These are summarized here for
    reference:

    1. We tried to ray-trace the fastest and slowest neutrons of the subframes back to
       the point in time and space where they intersected, to get the converging point
       for all neutrons. This gives a different distance for each subframe, which is
       not really supported in the current implementation of transform_coords.

    2. A slightly modified version of idea 1. was to ray-trace back to the intersection
       point for each frame, and then compute a mean distance that would apply to all
       frames. We then re-traced all subframes back to this mean distance. This seems to
       give a worse estimate of the time-of-flight origin, probably because in a
       slightly faulty chopper setup (such as V20), we are not always guaranteed that
       the neutrons went through the openings they were meant to go through. We are
       however relatively sure which openings of the first chopper they went through.
    """
    sorted_frame = chopper_cascade.Frame(
        distance=clean_detector_frame.distance,
        subframes=sorted(
            clean_detector_frame.subframes, key=lambda x: x.start_time.min()
        ),
    )

    times = sorted_frame.subbounds()['time'].flatten(
        dims=['subframe', 'bound'], to='subframe'
    )
    # All times before the first subframe and after the last subframe should be
    # replaced by NaN. We add a large padding to make sure all events are covered.
    padding = sc.scalar(1e9, unit='s').to(unit=times.unit)
    low = times['subframe', 0] - padding
    high = times['subframe', -1] + padding
    times = sc.concat([low, times, high], 'subframe')

    # Propagate the subframes from the detector back to the position of the first
    # chopper (note that frame 0 is the pulse itself).
    at_first_chopper = sorted_frame.propagate_to(frames[1].distance)

    starts = sc.concat(
        [subframe.start_time for subframe in at_first_chopper.subframes], dim='subframe'
    )
    ends = sc.concat(
        [subframe.end_time for subframe in at_first_chopper.subframes], dim='subframe'
    )
    time_origins = 0.5 * (starts + ends)

    # We need to add nans between each crossing_time offsets for the bins before,
    # after, and between the subframes.
    sizes = time_origins.sizes
    sizes['subframe'] = 2 * sizes['subframe'] + 1
    shift = sc.full(sizes=sizes, value=math.nan, unit='s')
    shift['subframe', 1::2] = time_origins

    origin_time = sc.DataArray(
        shift.transpose(times.dims).copy(deep=True), coords={'subframe': times}
    )

    return TimeOfFlightOrigin(time=origin_time, distance=at_first_chopper.distance)


def unwrap_data(da: RawData, delta: DeltaFromWrapped) -> UnwrappedData:
    """
    Return the input data with unwrapped time offset and pulse time.

    The time offset is the time since the start of the frame emitting the neutron.
    The pulse time is the time of the pulse that emitted the neutron.

    Parameters
    ----------
    da :
        The input data.
    delta :
        The positive delta that needs to be added to the input time offsets to unwrap
        them. The same delta is also subtracted from the input time zero to obtain the
        pulse time.
    """
    if da.bins is not None:
        da = da.copy(deep=False)
        da.data = sc.bins(**da.bins.constituents)
        da.bins.coords['time_offset'] = da.bins.coords['event_time_offset'] + delta
        if 'event_time_zero' in da.bins.coords:
            coord = da.bins.coords['event_time_zero']
        else:
            # Bin edges are now invalid so we pop them
            coord = da.coords.pop('event_time_zero')
        da.bins.coords['pulse_time'] = coord - delta.to(
            unit=elem_unit(coord), dtype='int64'
        )
    else:
        # 'time_of_flight' is the name in, e.g., NXmonitor
        da = da.transform_coords(
            time_offset=lambda time_of_flight: time_of_flight + delta,
            keep_inputs=False,
        )
        # Generally the coord is now not ordered, might want to cut, swap, and concat,
        # but it is unclear what to do with the split bin in the middle and how to
        # join the two halves. The end of the last bin generally does not match the
        # begin of the first bin, there may be a significant gap (which we could fill
        # with NaNs or zeros), but there could also be a small overlap.
    return UnwrappedData(da)


def to_time_of_flight(
    da: UnwrappedData, origin: TimeOfFlightOrigin, ltotal: Ltotal
) -> TofData:
    """
    Return the input data with 'tof', 'time_zero', and corrected 'Ltotal' coordinates.

    The 'tof' coordinate is the time-of-flight of the neutron, i.e., the time
    difference between the time of arrival of the neutron at the detector, and the
    time of the neutron at the source or source-chopper. The 'time_zero' coordinate is
    the time of the neutron at the source or source-chopper.

    It is critical to compute the 'time_zero' coordinate precisely by applying the
    reverse of the offset to time-of-flight, since this may be used to compute a
    precise time at the sample position. This is important for sample environments
    with highly time-dependent properties, since precise event-filtering based on
    sample environment data may be required.
    """
    time_offset = (
        da.coords['time_offset'] if da.bins is None else da.bins.coords['time_offset']
    )
    delta = origin.time
    if isinstance(delta, sc.DataArray):
        # Will raise if subframes overlap, since coord for lookup table must be sorted
        delta = sc.lookup(delta, dim='subframe')[time_offset]
    if da.bins is not None:
        da = da.copy(deep=False)
        da.data = sc.bins(**da.bins.constituents)
        da.bins.coords['tof'] = time_offset - delta
        da.bins.coords['time_zero'] = da.bins.coords['pulse_time'] + delta.to(
            unit=elem_unit(da.bins.coords['pulse_time']), dtype='int64'
        )
    else:
        da = da.transform_coords(
            tof=lambda time_offset: time_offset - delta, keep_inputs=False
        )
    if (existing := da.coords.get('Ltotal')) is not None:
        if not sc.identical(existing, ltotal):
            raise ValueError(
                f"Ltotal {existing} in data does not match Ltotal {ltotal} "
                "used for calculating time-of-flight."
            )

    da.coords['Ltotal'] = ltotal - origin.distance
    return TofData(da)


_common_providers = (
    chopper_cascade_frames,
    frame_at_detector,
    frame_bounds,
    frame_period,
    pulse_wrapped_time_offset,
    source_chopper,
    offset_from_wrapped,
    unwrap_data,
)

_non_skipping = (frame_wrapped_time_offset,)
_skipping = (frame_wrapped_time_offset_pulse_skipping, pulse_offset, time_zero)

_wfm = (
    maybe_clip_detector_subframes,
    time_offset,
    time_of_flight_origin_wfm,
)
_non_wfm = (time_of_flight_origin_from_chopper,)


def time_of_flight_providers() -> tuple[Callable]:
    """
    Return the providers computing the time-of-flight and time-zero coordinates.
    """
    return (to_time_of_flight,)


def time_of_flight_origin_from_choppers_providers(wfm: bool = False):
    """
    Return the providers for computing the time-of-flight origin via the chopper
    cascade.

    Parameters
    ----------
    wfm :
        If True, the data is assumed to be from a WFM instrument.
    """
    wfm = _wfm if wfm else _non_wfm
    _common = (source_chopper, frame_at_detector)
    return _common + wfm


def unwrap_providers(pulse_skipping: bool = False):
    """
    Return the list of providers for unwrapping frames.

    Parameters
    ----------
    pulse_skipping :
        If True, the pulse-skipping mode is assumed.
    """
    skipping = _skipping if pulse_skipping else _non_skipping
    return _common_providers + skipping



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""File writer and reader for XYE files."""

import io
from pathlib import Path

import numpy as np
import scipp as sc

from ..logging import get_logger


class GenerateHeaderType:
    def __repr__(self) -> str:
        return 'GenerateHeader'


GenerateHeader = GenerateHeaderType()


def save_xye(
    fname: str | Path | io.TextIOBase,
    da: sc.DataArray,
    *,
    coord: str | None = None,
    header: str | GenerateHeaderType = GenerateHeader,
) -> None:
    """Write a data array to an XYE file.

    The input must be 1-dimensional, have variances, and at least one coordinate.
    It is possible to select which coordinate gets written with the ``coord`` argument.

    The data are written as an ASCII table with columns X, Y, E, where

    - X: coordinate of the data array,
    - Y: data values of the data array,
    - E: standard deviations corresponding to Y.

    This format is lossy, coordinates other than X, attributes,
    and masks are not written and the coordinate name, dimension name, and
    units of the input are lost.
    To improve the situation slightly, ``save_xye`` writes a basic header by default.
    All lines in the header are prefixed with ``#``.

    Parameters
    ----------
    fname:
        Name or file handle of the output file.
    da:
        1-dimensional data to write.
    coord:
        Coordinate name of ``da`` to write.
        If omitted and ``da`` has only one coordinate, that coordinate is used.
        If omitted and ``da`` has multiple coordinates, attempts to use a coordinate
        with the same name as ``da.dim``.
        If that does not exist, raise an error.
    header:
        String to write at the beginning of the file.
        A simple table header gets generated automatically by default.
        Set ``header=''`` to prevent this.

    See Also
    --------
    scippneutron.io.xye.load_xye:
        Function to load XYE files.
    """
    if da.variances is None:
        raise sc.VariancesError(
            'Cannot save data to XYE file because it has no variances.'
        )
    if da.ndim != 1:
        raise sc.DimensionError(
            'Cannot save data to XYE file because it is not one-dimensional. '
            f'It has dimensions {da.dims}'
        )
    if da.masks:
        raise ValueError(
            'Cannot save data to XYE file because it has masks: '
            f'{list(da.masks.keys())}'
        )
    if len(da.coords) == 0:
        raise ValueError('Cannot save data to XYE file because it has no coordinates.')
    coord = _deduce_coord(da) if coord is None else coord
    if da.coords.is_edges(coord):
        raise sc.CoordError(
            'Cannot save data with bin-edges to XYE file. '
            'Compute bin-centers before calling save_xye. '
            'Use, e.g., scipp.midpoints for linearly spaced bins.'
        )
    to_save = np.c_[da.coords[coord].values, da.values, np.sqrt(da.variances)]
    if header is GenerateHeader:
        header = _generate_xye_header(da, coord)

    get_logger().info(
        "Saving data with unit %s and coordinate '%s' to XYE file %s",
        da.unit,
        coord,
        fname,
    )
    np.savetxt(fname, to_save, delimiter=' ', header=header)


def load_xye(
    fname: str | Path | io.TextIOBase,
    *,
    dim: str,
    unit: sc.Unit | str | None,
    coord_unit: sc.Unit | str | None,
    coord: str | None = None,
) -> sc.DataArray:
    """Read a data array from an XYE file.

    See :func:`scippneutron.io.xye.save_xye` for a description of the file format.

    Since XYE files are lossy, some metadata must be provided manually when calling
    this function.

    Parameters
    ----------
    fname:
        Name or file handle of the input file.
    dim:
        Dimension of the returned data.
    unit:
        Unit of the returned data array.
    coord_unit:
        Unit of the coordinate of the returned data array.
    coord:
        Coordinate name of the returned data.
        Defaults to the value of ``dim``.

    Returns
    -------
    da:
        Data array read from the file.

    See Also
    --------
    scippneutron.io.xye.save_xye:
        Function to write XYE files.
    """
    coord = dim if coord is None else coord
    loaded = np.loadtxt(fname, delimiter=' ', unpack=True)
    if loaded.ndim == 1:
        loaded = loaded[:, np.newaxis]
    return sc.DataArray(
        sc.array(dims=[dim], values=loaded[1], variances=loaded[2] ** 2, unit=unit),
        coords={coord: sc.array(dims=[dim], values=loaded[0], unit=coord_unit)},
    )


def _generate_xye_header(da: sc.DataArray, coord: str) -> str:
    def format_unit(unit):
        return f'[{unit}]' if unit is not None else ''

    c = f'{coord} {format_unit(da.coords[coord].unit)}'
    y = f'Y {format_unit(da.unit)}'
    e = f'E {format_unit(da.unit)}'
    # Widths are for the default format of `0.0`
    return f'{c:22} {y:24} {e:24}'


def _deduce_coord(da: sc.DataArray) -> str:
    if len(da.coords) == 1:
        return next(iter(da.coords))
    if len(da.coords) > 1 and da.dim not in da.coords:
        raise ValueError(
            'Cannot deduce which coordinate to save because the data has more '
            f'than one and no dimension-coordinate (dim={da.dim}): '
            f'{list(da.coords.keys())}'
        )
    return da.dim



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Any

import matplotlib
import pytest
import scipp as sc

pytest.register_assert_rewrite('scipp.testing.assertions')


def pytest_assertrepr_compare(op: str, left: Any, right: Any) -> list[str]:
    if isinstance(left, sc.Unit) and isinstance(right, sc.Unit):
        return [f'Unit({left}) {op} Unit({right})']
    if isinstance(left, sc.DType) or isinstance(right, sc.DType):
        return [f'{left!r} {op} {right!r}']


@pytest.fixture
def _use_ipympl():
    """
    Use ipympl interactive backend for matplotlib.
    Close figures when done, and reset matplotlib defaults.
    """
    matplotlib.use('module://ipympl.backend_nbagg')
    yield
    matplotlib.rcdefaults()
    matplotlib.use('Agg')



from dataclasses import dataclass

import scipp as sc

from scippneutron.atoms import ScatteringParams, reference_wavelength


@dataclass
class Material:
    scattering_params: ScatteringParams
    effective_sample_number_density: sc.Variable
    '''Density of the sample in number of formulas per cubic angstrom'''

    def attenuation_coefficient(self, wavelength: sc.Variable) -> sc.Variable:
        '''Computes marginal attenuation per distance for
        the given neutron wavelength.'''
        return self.effective_sample_number_density * (
            self.scattering_params.total_scattering_cross_section
            + (
                self.scattering_params.absorption_cross_section
                * (wavelength / reference_wavelength().to(unit=wavelength.unit))
            ).to(unit=self.scattering_params.total_scattering_cross_section.unit)
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen

"""Graphs for coordinate transformations.

All graphs are defined in terms of the functions in the parent module
:mod:`scippneutron.conversion`.
See there for definitions of the individual conversions.

Typically, multiple graphs need to be combined for a full transformation.
For example:

    >>> from scippneutron.conversion import graph
    >>> wavelength_graph = {**graph.beamline.beamline(scatter=True),
    ...                     **graph.tof.elastic_wavelength(start='tof')}

The `user guide <../../user-guide/coordinate-transformations.rst>`_ gives
more examples.
"""

from . import beamline, tof

__all__ = ['beamline', 'tof']



"""Make plots of the computed times for disk choppers."""

from dataclasses import replace
from itertools import cycle
from pathlib import Path

import matplotlib.pyplot as plt
import scipp as sc
from matplotlib.patches import Rectangle
from numpy import pi

from scippneutron.chopper import DiskChopper

OUT_DIR = Path(__file__).resolve().parent.parent / '_static' / 'chopper'

GENERAL_STYLE = {
    'axes.labelsize': 'xx-large',
    'axes.titlesize': 'xx-large',
    'xtick.major.size': 5,
    'xtick.labelsize': 'x-large',
    'ytick.major.size': 5,
    'ytick.labelsize': 'x-large',
}


def set_light_style():
    plt.style.use(
        [
            'default',
            GENERAL_STYLE,
            {
                # High jack the color cycle to get theme-specific grid colors.
                'axes.prop_cycle': plt.cycler(
                    'color', ['#1f77b4', '#ff7f0e', '0.2', '0.5']
                ),
            },
        ]
    )


def set_dark_style():
    plt.style.use(
        [
            'dark_background',
            GENERAL_STYLE,
            {
                # Colors from the pydata-sphinx-theme.
                'axes.edgecolor': 'c3d6dd',
                'axes.facecolor': '#14181e',
                'axes.labelcolor': 'c3d6dd',
                'figure.edgecolor': '#14181e',
                'figure.facecolor': '#14181e',
                'savefig.edgecolor': '#14181e',
                'savefig.facecolor': '#14181e',
                'text.color': 'c3d6dd',
                # High jack the color cycle to get theme-specific grid colors.
                'axes.prop_cycle': plt.cycler(
                    'color', ['#8dd3c7', '#feffb3', '0.8', '0.5']
                ),
            },
        ]
    )


def plot_axes_lines(ax, x_ticks, y_ticks):
    for x in x_ticks:
        ax.axvline(x=x, color='C2' if x == 0 else 'C3', ls=':')
    for y in y_ticks:
        ax.axhline(y=y, color='C2' if y == 0 else 'C3', ls=':')


def plot_t_vs_angle_anticlockwise(ax, ch):
    theta_tilde = ch.beam_position.value
    omega = ch.angular_frequency.value

    theta = sc.linspace('theta', -pi / 2, 3.2 * pi, 100, unit='rad')
    time = ch.time_offset_angle_at_beam(angle=theta)
    time_wrapped = time % (1 / ch.frequency)

    x_ticks = [0.0, theta_tilde, 2 * pi, 2 * pi + theta_tilde]
    x_tick_labels = ['0', r'$\tilde{\theta}$', r'$2\pi$', r'$2\pi+\tilde{\theta}$']

    y_ticks = [0.0, theta_tilde / omega, 2 * pi / omega, (2 * pi + theta_tilde) / omega]
    y_tick_labels = [
        '0',
        r'$\frac{\tilde{\theta}}{\omega}$',
        r'$\frac{2\pi}{\omega}$',
        r'$\frac{2\pi+\tilde{\theta}}{\omega}$',
    ]

    ax.add_patch(
        Rectangle(
            (0, theta_tilde / omega), 2 * pi, 2 * pi / omega, facecolor='#80808040'
        )
    )
    plot_axes_lines(ax, x_ticks, y_ticks)
    ax.plot(theta.values[[0, -1]], time.values[[0, -1]], linewidth=2)
    ax.plot(theta.values, time_wrapped.values, linewidth=2)
    ax.set_xlabel(rf'$\theta$ [{theta.unit}]')
    ax.set_ylabel(rf'$\Delta t_g(\theta)$ [{time.unit}]')
    ax.set_xticks(x_ticks, x_tick_labels)
    ax.set_yticks(y_ticks, y_tick_labels)


def plot_t_vs_angle_clockwise(ax, ch):
    theta_tilde = ch.beam_position.value
    omega = abs(ch.angular_frequency.value)

    theta = sc.linspace('theta', -pi / 2, 3.2 * pi, 100, unit='rad')
    time = ch.time_offset_angle_at_beam(angle=theta)
    time_wrapped = time % (1 / abs(ch.frequency))

    x_ticks = [0, theta_tilde, 2 * pi, 2 * pi + theta_tilde]
    x_tick_labels = ['0', r'$\tilde{\theta}$', r'$2\pi$', r'$2\pi + \tilde{\theta}$']

    y_ticks = [
        -theta_tilde / omega,
        0.0,
        (2 * pi - theta_tilde) / omega,
        2 * pi / omega,
    ]
    y_tick_labels = [
        r'$-\frac{\tilde{\theta}}{|\omega|}$',
        '0',
        r'$\frac{2\pi - \tilde{\theta}}{|\omega|}$',
        r'$\frac{2\pi}{|\omega|}$',
    ]

    ax.add_patch(
        Rectangle(
            (0, -theta_tilde / omega), 2 * pi, 2 * pi / omega, facecolor='#80808040'
        )
    )
    plot_axes_lines(ax, x_ticks, y_ticks)
    ax.plot(theta.values[[0, -1]], time.values[[0, -1]], linewidth=2)
    ax.plot(theta.values, time_wrapped.values, linewidth=2)
    ax.set_xlabel(r'$\theta$ [rad]')
    ax.set_ylabel(r'$\Delta t_g(\theta)$')
    ax.set_xticks(x_ticks, x_tick_labels)
    ax.set_yticks(y_ticks, y_tick_labels)


def plot_t_vs_angle(ch, name):
    fig, axs = plt.subplots(1, 2, layout='constrained', figsize=(11, 5))

    axs[0].set_title('Clockwise')
    plot_t_vs_angle_clockwise(axs[0], replace(ch, frequency=-ch.frequency))

    axs[1].set_title('Anticlockwise')
    assert not ch.is_clockwise  # noqa: S101
    plot_t_vs_angle_anticlockwise(axs[1], ch)

    fig.savefig(OUT_DIR.joinpath(name).with_suffix('.svg'))


def plot_openings_for_multiple(ax, y, ch, pulse_frequency, n, clockwise):
    ch = replace(ch, frequency=n * pulse_frequency * (-1 if clockwise else 1))
    open_times = ch.time_offset_open(pulse_frequency=pulse_frequency).to(unit='s')
    close_times = ch.time_offset_close(pulse_frequency=pulse_frequency).to(unit='s')
    for i, o, c in zip(cycle((0, 1)), open_times, close_times):
        ax.plot([o.value, c.value], [y, y], lw=10, c=f'C{i}', solid_capstyle='butt')


def plot_pulses(ax, period, clockwise):
    if clockwise:
        x = [-period, 0, period, 2 * period]
        ax.plot([-period, 0], [0, 0], lw=10, c='C3', solid_capstyle='butt')
        ax.plot([0, period], [0, 0], lw=10, c='C2', solid_capstyle='butt')
        ax.plot([period, period * 2], [0, 0], lw=10, c='C3', solid_capstyle='butt')
        ax.set_xticks(
            x, [r'$T_0 - \Delta T$', '$T_0$', r'$T_0 + \Delta T$', r'$T_0 + 2\Delta T$']
        )
    else:
        x = [0, period, 2 * period, 3 * period]
        ax.plot([0, period], [0, 0], lw=10, c='C2', solid_capstyle='butt')
        ax.plot(
            [period, period * 2, period * 3],
            [0, 0, 0],
            lw=10,
            c='C3',
            solid_capstyle='butt',
        )
        ax.set_xticks(
            x,
            ['$T_0$', r'$T_0 + \Delta T$', r'$T_0 + 2\Delta T$', r'$T_0 + 3\Delta T$'],
        )
    ax.plot(x, [0] * len(x), ls='', markersize=20, marker='|', c='C2')


def plot_openings_clockwise(ax, ch, pulse_frequency, clockwise):
    pulse_frequency = pulse_frequency.to(unit='Hz')
    period = sc.reciprocal(pulse_frequency).value
    plot_pulses(ax, period, clockwise)

    ns = [0.5, 1, 2]
    ys = []
    for i, n in enumerate(ns, 1):
        ys.append(0.1 * i)
        plot_openings_for_multiple(ax, ys[-1], ch, pulse_frequency, n, clockwise)
    ax.set_yticks(ys, list(map(str, ns)))

    ax.set_xlabel('$t_g$')
    ax.set_ylabel('frequency ratio')
    ax.set_ylim((-0.04, 0.34))


def plot_openings(ch, name, pulse_frequency):
    ch = replace(
        ch,
        beam_position=sc.scalar(5, unit='deg'),
        phase=sc.scalar(360 + 50, unit='deg'),
    )

    fig, axs = plt.subplots(1, 2, layout='constrained', figsize=(11, 2.5))

    axs[0].set_title('Clockwise')
    plot_openings_clockwise(axs[0], ch, pulse_frequency, True)

    axs[1].set_title('Anticlockwise')
    plot_openings_clockwise(axs[1], ch, pulse_frequency, False)

    fig.savefig(OUT_DIR.joinpath(name).with_suffix('.svg'))


def main() -> None:
    OUT_DIR.mkdir(exist_ok=True)
    ch = DiskChopper(
        axle_position=sc.vector([0.0, 0.0, 0.0], unit='m'),
        frequency=sc.scalar(2.3, unit='Hz'),
        beam_position=sc.scalar(2.5, unit='rad'),
        phase=sc.scalar(0.0, unit='rad'),
        slit_begin=sc.array(dims=['slit'], values=[10.0, 150.0], unit='deg'),
        slit_end=sc.array(dims=['slit'], values=[70.0, 280.0], unit='deg'),
    )
    pulse_frequency = sc.scalar(10.0, unit='Hz')

    set_light_style()
    # plot_t_vs_angle(ch, 'disk-chopper-time-curve')
    # plot_openings(ch, 'disk-chopper-openings', pulse_frequency)
    set_dark_style()
    # plot_t_vs_angle(ch, 'disk-chopper-time-curve-dark')
    plot_openings(ch, 'disk-chopper-openings-dark', pulse_frequency)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402

"""Neutron scattering toolkit built using scipp for Data Reduction.

ScippNeutron is a generic (as in 'usable by different facilities')
package for data processing in neutron scattering.
It provides coordinate transformations, file I/O, and technique-specific tools.

See the online documentation for user guides and the API reference:
https://scipp.github.io/scippneutron/
"""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .beamline_components import (
    position,
    source_position,
    sample_position,
    incident_beam,
    scattered_beam,
    Ltotal,
    L1,
    L2,
    two_theta,
)
from .core import convert
from .mantid import (
    from_mantid,
    array_from_mantid,
    to_mantid,
    load_with_mantid,
    load,
    fit,
)
from .instrument_view import instrument_view
from .masking import MaskingTool
from . import atoms
from . import chopper
from . import io

del importlib

__all__ = [
    "position",
    "source_position",
    "sample_position",
    "incident_beam",
    "scattered_beam",
    "Ltotal",
    "L1",
    "L2",
    "two_theta",
    "convert",
    "from_mantid",
    "array_from_mantid",
    "to_mantid",
    "io",
    "load_with_mantid",
    "load",
    "fit",
    "instrument_view",
    "atoms",
    "chopper",
    "MaskingTool",
]



import pytest
import scipp as sc
from scipp.testing import assert_identical

from scippneutron.tof import unwrap


def test_to_time_of_flight_raises_if_Ltotal_incompatible() -> None:
    unwrapped = sc.DataArray(
        data=sc.scalar(1.0, unit='counts'),
        coords={
            'time_offset': sc.scalar(0.3, unit='s'),
            'Ltotal': sc.scalar(3.0, unit='m'),
        },
    )
    origin = unwrap.TimeOfFlightOrigin(
        time=sc.scalar(0.1, unit='s'), distance=sc.scalar(1.0, unit='m')
    )
    ltotal = sc.scalar(2.0, unit='m')
    with pytest.raises(ValueError, match='Ltotal'):
        unwrap.to_time_of_flight(unwrapped, origin=origin, ltotal=ltotal)


def test_to_time_of_flight_subtracts_from_time_offset() -> None:
    unwrapped = sc.DataArray(
        data=sc.scalar(1.0, unit='counts'),
        coords={
            'time_offset': sc.scalar(3.0, unit='s'),
            'Ltotal': sc.scalar(3.0, unit='m'),
        },
    )
    origin = unwrap.TimeOfFlightOrigin(
        time=sc.scalar(1.0, unit='s'), distance=sc.scalar(1.0, unit='m')
    )
    ltotal = sc.scalar(3.0, unit='m')
    result = unwrap.to_time_of_flight(unwrapped, origin=origin, ltotal=ltotal)
    assert_identical(result.coords['tof'], sc.scalar(2.0, unit='s'))


def test_to_time_of_flight_event_mode() -> None:
    content = sc.DataArray(
        data=sc.array(dims=['event'], values=[1.0, 2.0], unit='counts'),
        coords={
            'time_offset': sc.array(dims=['event'], values=[3.0, 4.0], unit='s'),
            'pulse_time': sc.array(dims=['event'], values=[10.0, 20.0], unit='s'),
        },
    )
    unwrapped = sc.DataArray(data=sc.bins(begin=sc.index(0), dim='event', data=content))
    origin = unwrap.TimeOfFlightOrigin(
        time=sc.scalar(1.0, unit='s'), distance=sc.scalar(1.0, unit='m')
    )
    ltotal = sc.scalar(3.0, unit='m')
    result = unwrap.to_time_of_flight(unwrapped, origin=origin, ltotal=ltotal)

    assert_identical(result.coords['Ltotal'], sc.scalar(3.0 - 1.0, unit='m'))
    assert_identical(
        result.value.coords['tof'],
        sc.array(dims=['event'], values=[2.0, 3.0], unit='s'),
    )
    assert_identical(
        result.value.coords['time_zero'],
        sc.array(dims=['event'], values=[11.0, 21.0], unit='s'),
    )
